# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from oslo_log import log as logging
import requests
import six
from six.moves.urllib.parse import quote

from neutron_lib.api.definitions import portbindings

from networking_arista._i18n import _, _LI, _LW, _LE
from networking_arista.common import constants as const
from networking_arista.common import exceptions as arista_exc
from networking_arista.common import utils
from networking_arista.ml2.rpc.base import AristaRPCWrapperBase

LOG = logging.getLogger(__name__)


class AristaRPCWrapperEapi(AristaRPCWrapperBase):
    def __init__(self):
        super(AristaRPCWrapperEapi, self).__init__()
        # The cli_commands dict stores the mapping between the CLI command key
        # and the actual CLI command.
        self.cli_commands = {
            'resource-pool': [],
            'features': {},
        }

    def _send_eapi_req(self, cmds, commands_to_log=None):
        # This method handles all EAPI requests (using the requests library)
        # and returns either None or response.json()['result'] from the EAPI
        # request.
        #
        # Exceptions related to failures in connecting/ timeouts are caught
        # here and logged. Other unexpected exceptions are logged and raised

        request_headers = {}
        request_headers['Content-Type'] = 'application/json'
        request_headers['Accept'] = 'application/json'
        url = self._api_host_url(host=self._server_ip)

        params = {}
        params['timestamps'] = "false"
        params['format'] = "json"
        params['version'] = 1
        params['cmds'] = cmds

        data = {}
        data['id'] = "Arista ML2 driver"
        data['method'] = "runCmds"
        data['jsonrpc'] = "2.0"
        data['params'] = params

        response = None

        try:
            # NOTE(pbourke): shallow copy data and params to remove sensitive
            # information before logging
            log_data = dict(data)
            log_data['params'] = dict(params)
            log_data['params']['cmds'] = commands_to_log or cmds
            msg = (_('EAPI request to %(ip)s contains %(cmd)s') %
                   {'ip': self._server_ip, 'cmd': json.dumps(log_data)})
            LOG.info(msg)
            response = requests.post(url, timeout=self.conn_timeout,
                                     verify=False, data=json.dumps(data))
            LOG.info(_LI('EAPI response contains: %s'), response.json())
            try:
                return response.json()['result']
            except KeyError:
                if response.json()['error']['code'] == 1002:
                    for data in response.json()['error']['data']:
                        if type(data) == dict and 'errors' in data:
                            if const.ERR_CVX_NOT_LEADER in data['errors'][0]:
                                msg = six.text_type("%s is not the master" % (
                                                    self._server_ip))
                                LOG.info(msg)
                                return None

                msg = "Unexpected EAPI error"
                LOG.info(msg)
                raise arista_exc.AristaRpcError(msg=msg)
        except requests.exceptions.ConnectionError:
            msg = (_('Error while trying to connect to %(ip)s') %
                   {'ip': self._server_ip})
            LOG.warning(msg)
            return None
        except requests.exceptions.ConnectTimeout:
            msg = (_('Timed out while trying to connect to %(ip)s') %
                   {'ip': self._server_ip})
            LOG.warning(msg)
            return None
        except requests.exceptions.Timeout:
            msg = (_('Timed out during an EAPI request to %(ip)s') %
                   {'ip': self._server_ip})
            LOG.warning(msg)
            return None
        except requests.exceptions.InvalidURL:
            msg = (_('Ignore attempt to connect to invalid URL %(ip)s') %
                   {'ip': self._server_ip})
            LOG.warning(msg)
            return None
        except ValueError:
            LOG.info("Ignoring invalid JSON response")
            return None
        except Exception as error:
            msg = six.text_type(error)
            LOG.warning(msg)
            raise

    def check_vlan_type_driver_commands(self):
        """Checks the validity of CLI commands for Arista's VLAN type driver.

           This method tries to execute the commands used exclusively by the
           arista_vlan type driver and stores the commands if they succeed.
        """
        cmd = ['show openstack resource-pool vlan region %s uuid'
               % self.region]
        try:
            self._run_eos_cmds(cmd)
            self.cli_commands['resource-pool'] = cmd
        except arista_exc.AristaRpcError:
            self.cli_commands['resource-pool'] = []
            LOG.warning(
                _LW("'resource-pool' command '%s' is not available on EOS"),
                cmd)

    def get_vlan_assignment_uuid(self):
        """Returns the UUID for the region's vlan assignment on CVX

        :returns: string containing the region's vlan assignment UUID
        """
        vlan_uuid_cmd = self.cli_commands['resource-pool']
        if vlan_uuid_cmd:
            return self._run_eos_cmds(commands=vlan_uuid_cmd)[0]
        return None

    def get_vlan_allocation(self):
        """Returns the status of the region's VLAN pool in CVX

        :returns: dictionary containg the assigned, allocated and available
                  VLANs for the region
        """
        if not self.cli_commands['resource-pool']:
            LOG.warning(_('The version of CVX you are using does not support '
                          'arista VLAN type driver.'))
        else:
            cmd = ['show openstack resource-pools region %s' % self.region]
            command_output = self._run_eos_cmds(cmd)
            if command_output:
                regions = command_output[0]['physicalNetwork']
                if self.region in regions.keys():
                    return regions[self.region]['vlanPool']['default']
        return {'assignedVlans': '',
                'availableVlans': '',
                'allocatedVlans': ''}

    def _run_eos_cmds(self, commands, commands_to_log=None):
        """Execute/sends a CAPI (Command API) command to EOS.

        In this method, list of commands is appended with prefix and
        postfix commands - to make is understandble by EOS.

        :param commands : List of command to be executed on EOS.
        :param commands_to_log : This should be set to the command that is
                                 logged. If it is None, then the commands
                                 param is logged.
        """

        # Always figure out who is master (starting with the last known val)
        try:
            if self._get_eos_master() is None:
                msg = "Failed to identify CVX master"
                self.set_cvx_unavailable()
                raise arista_exc.AristaRpcError(msg=msg)
        except Exception:
            self.set_cvx_unavailable()
            raise

        self.set_cvx_available()
        log_cmds = commands
        if commands_to_log:
            log_cmds = commands_to_log

        LOG.info(_LI('Executing command on Arista EOS: %s'), log_cmds)
        # this returns array of return values for every command in
        # full_command list
        try:
            response = self._send_eapi_req(cmds=commands,
                                           commands_to_log=log_cmds)
            if response is None:
                # Reset the server as we failed communicating with it
                self._server_ip = None
                self.set_cvx_unavailable()
                msg = "Failed to communicate with CVX master"
                raise arista_exc.AristaRpcError(msg=msg)
            return response
        except arista_exc.AristaRpcError:
            raise

    def _build_command(self, cmds, sync=False):
        """Build full EOS's openstack CLI command.

        Helper method to add commands to enter and exit from openstack
        CLI modes.

        :param cmds: The openstack CLI commands that need to be executed
                     in the openstack config mode.
        :param sync: This flags indicates that the region is being synced.
        """

        region_cmd = 'region %s' % self.region
        if sync:
            region_cmd = self.cli_commands[const.CMD_REGION_SYNC]

        full_command = [
            'enable',
            'configure',
            'cvx',
            'service openstack',
            region_cmd,
        ]
        full_command.extend(cmds)
        return full_command

    def _run_openstack_cmds(self, commands, commands_to_log=None, sync=False):
        """Execute/sends a CAPI (Command API) command to EOS.

        In this method, list of commands is appended with prefix and
        postfix commands - to make is understandble by EOS.

        :param commands : List of command to be executed on EOS.
        :param commands_to_logs : This should be set to the command that is
                                  logged. If it is None, then the commands
                                  param is logged.
        :param sync: This flags indicates that the region is being synced.
        """

        full_command = self._build_command(commands, sync=sync)
        if commands_to_log:
            full_log_command = self._build_command(commands_to_log, sync=sync)
        else:
            full_log_command = None
        return self._run_eos_cmds(full_command, full_log_command)

    def _get_eos_master(self):
        # Use guarded command to figure out if this is the master
        cmd = ['show openstack agent uuid']

        cvx = self._get_cvx_hosts()
        # Identify which EOS instance is currently the master
        for self._server_ip in cvx:
            try:
                response = self._send_eapi_req(cmds=cmd, commands_to_log=cmd)
                if response is not None:
                    return self._server_ip
                else:
                    continue  # Try another EOS instance
            except Exception:
                raise

        # Couldn't find an instance that is the leader and returning none
        self._server_ip = None
        msg = "Failed to reach the CVX master"
        LOG.error(msg)
        return None

    def _api_host_url(self, host=""):
        return ('https://%s:%s@%s/command-api' %
                (quote(self._api_username()),
                 quote(self._api_password()),
                 host))

    def get_baremetal_physnet(self, context):
        """Returns dictionary which contains mac to hostname mapping"""
        port = context.current
        host_id = context.host
        cmd = ['show network physical-topology hosts']
        try:
            response = self._run_eos_cmds(cmd)
            binding_profile = port.get(portbindings.PROFILE, {})
            link_info = binding_profile.get('local_link_information', [])
            for link in link_info:
                switch_id = link.get('switch_id')
                for host in response[0]['hosts'].values():
                    if switch_id == host['name']:
                        physnet = utils.physnet(host['hostname'])
                        LOG.debug("get_physical_network: Physical Network for "
                                  "%(host)s is %(physnet)s",
                                  {'host': host_id, 'physnet': physnet})
                        return physnet
            LOG.debug("Physical network not found for %(host)s",
                      {'host': host_id})
        except Exception as exc:
            LOG.error(_LE('command %(cmd)s failed with '
                      '%(exc)s'), {'cmd': cmd, 'exc': exc})
        return None

    def get_host_physnet(self, context):
        """Returns dictionary which contains physical topology information

        for a given host_id
        """
        host_id = utils.hostname(context.host)
        cmd = ['show network physical-topology neighbors']
        try:
            response = self._run_eos_cmds(cmd)
            # Get response for 'show network physical-topology neighbors'
            # command
            neighbors = response[0]['neighbors']
            for neighbor in neighbors:
                if host_id in neighbor:
                    physnet = utils.physnet(
                        neighbors[neighbor]['toPort'][0]['hostname'])
                    LOG.debug("get_physical_network: Physical Network for "
                              "%(host)s is %(physnet)s", {'host': host_id,
                                                          'physnet': physnet})
                    return physnet
            LOG.debug("Physical network not found for %(host)s",
                      {'host': host_id})
        except Exception as exc:
            LOG.error(_LE('command %(cmd)s failed with '
                      '%(exc)s'), {'cmd': cmd, 'exc': exc})
        return None



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib.plugins import directory

from networking_arista.tests.unit.ml2 import ml2_test_base
from networking_arista.tests.unit import utils


class SecurityGroupTestBase(ml2_test_base.MechTestBase):

    def get_additional_service_plugins(self):
        p = super(SecurityGroupTestBase, self).get_additional_service_plugins()
        p.update({'arista_security_group_plugin': 'arista_security_group'})
        return p

    def setUp(self):
        super(SecurityGroupTestBase, self).setUp()
        self.arista_sec_gp_plugin = directory.get_plugin(
            'arista_security_group')
        self.switch1 = utils.MockSwitch()
        self.switch2 = utils.MockSwitch()
        self.switches = {'TOR1': self.switch1,
                         'TOR2': self.switch2}
        self.arista_sec_gp_plugin._switches = self.switches
        self.arista_sec_gp_plugin._port_group_info['TOR1'] = {
            'Ethernet1': {'interfaceMembership': ''},
            'Ethernet2': {'interfaceMembership': ''}}
        self.arista_sec_gp_plugin._port_group_info['TOR2'] = {
            'Ethernet1': {'interfaceMembership': ''},
            'Ethernet2': {'interfaceMembership': ''}}

    def create_port_channel(self, switch, interface, pc_name):
        intf_info = self.arista_sec_gp_plugin._port_group_info[switch]
        intf_info[interface]['interfaceMembership'] = 'Member of %s' % pc_name

    def create_sg_rule(self, direction, proto, cidr, range_min=None,
                       range_max=None, ethertype='IPv4', default=True,
                       sg_id=None):
        if sg_id is None:
            sec_group = {'security_group':
                         {'name': 'sg1',
                          'tenant_id': 't1',
                          'description': ''}}
            grp = self.plugin.create_security_group(self.context, sec_group,
                                                    default_sg=default)
            sg_id = grp['id']
        for switch in self.switches.values():
            switch.clear_received_commands()
        rule = {'security_group_rule':
                {'direction': direction,
                 'ethertype': ethertype,
                 'protocol': proto,
                 'remote_ip_prefix': cidr,
                 'port_range_min': range_min,
                 'port_range_max': range_max,
                 'security_group_id': sg_id,
                 'remote_group_id': None,
                 'tenant_id': 't1'}}
        rule = self.plugin.create_security_group_rule(self.context, rule)
        return sg_id, rule



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from neutron_lib import constants as n_const
from neutron_lib.services.trunk import constants as t_const
from oslo_log import log as logging

from networking_arista.common import constants as a_const
from networking_arista.common import db_lib
from networking_arista.common import exceptions as arista_exc
from networking_arista.common import utils


LOG = logging.getLogger(__name__)


class AttributeFormatter(object):
    """Formats a single attribute of the CVX model based on the neutron model

    There are 4 elements to an AttributeFormatter:
    1. neutron_key - name of the key in the neutron model
    2. cvx_key - name of the key in the cvx model
    3. format(optional) - function to alter the value to be CVX compatible
    4. submodel(optional) - If the get_db_resources function queries multiple
       models, the name of the model that contains the neutron_key must be
       specified
    """

    def __init__(self, neutron_key, cvx_key, format=None, submodel=None):
        self.neutron_key = neutron_key
        self.cvx_key = cvx_key
        self.format = format or (
            lambda neutron_val, base_resource: neutron_val)
        self.submodel = submodel

    def transform(self, resource):
        base_resource = resource
        if self.submodel:
            resource = getattr(resource, self.submodel)
        return (self.cvx_key, self.format(resource[self.neutron_key],
                                          base_resource))


class AristaResourcesBase(object):
    """Tracks state of resources of one resource type on neutron and CVX

    An AristaResources class is responsible for:
    - tracking resources that have been provisioned in neutron
    - tracking resources that have been provisioned on CVX
    - creating and deleting resources on CVX to bring it in line with neutron
    - formatting neutron resources to be compatible with CVX's API
    - tracking the correct endpoint for CVX API calls

    In order to facilitate this each resource type should define:
    1. formatter - a list of AttributeFormatters to convert neutron attributes
       to models compatible with CVX's API
    2. id_key - the key in the CVX model that uniquely identifies the resource
    3. endpoint - format string for region resource endpoint
    4. get_db_resources - function that queries the neutron db for all
       resources of the resource type in question
    """

    formatter = [AttributeFormatter('id', 'id')]
    id_key = 'id'
    endpoint = 'region/%(region)s'

    def __init__(self, rpc):
        self.region = rpc.region
        self.rpc = rpc
        self.cvx_data_stale = True
        self.neutron_data_stale = True
        self.cvx_ids = set()
        self.neutron_resources = dict()

    def clear_cvx_data(self):
        self.cvx_data_stale = True
        self.cvx_ids = set()

    def clear_neutron_data(self):
        self.neutron_data_stale = True
        self.neutron_resources = dict()

    def clear_all_data(self):
        self.clear_cvx_data()
        self.clear_neutron_data()

    def update_neutron_resource(self, id, action):
        LOG.debug("%(pid)s Requesting %(action)s %(class)s resource %(id)s",
                  {'action': action, 'class': self.__class__.__name__,
                   'id': id, 'pid': os.getpid()})
        resource = self.get_db_resources(id)
        if action == a_const.FULL_SYNC:
            self.clear_neutron_data()
            LOG.info("%(pid)s Full sync resource %(class)s",
                     {'class': self.__class__.__name__,
                      'pid': os.getpid()})
            return

        assert(len(resource) <= 1)
        if resource:
            # Until we start using etcd, we need to unconditionally send the
            # create request because it might have been delete by another
            # worker.  We force this by removing the resource to our 'view' of
            # cvx resources
            self.force_resource_update(id)
            LOG.debug("%(pid)s Resource %(class)s %(id)s found, creating",
                      {'class': self.__class__.__name__, 'id': id,
                       'pid': os.getpid()})
            self._add_neutron_resource(resource[0])
        else:
            LOG.debug("%(pid)s Resource %(class)s %(id)s not found, deleting",
                      {'class': self.__class__.__name__, 'id': id,
                       'pid': os.getpid()})
            self._delete_neutron_resource(id)

    def _add_neutron_resource(self, resource):
        formatted_resource = self.format_for_create(resource)
        resource_id = list(formatted_resource.keys())[0]
        LOG.debug("%(pid)s %(class)s resource %(id)s added locally",
                  {'class': self.__class__.__name__,
                   'id': resource_id,
                   'pid': os.getpid()})
        # If the resource has changed, force a POST to CVX
        old_resource = self.neutron_resources.get(resource_id)
        if old_resource and old_resource != formatted_resource[resource_id]:
            LOG.debug("%(pid)s %(class)s resource %(id)s requires update",
                      {'class': self.__class__.__name__,
                       'id': resource_id,
                       'pid': os.getpid()})
            self.force_resource_update(resource_id)
        self.neutron_resources.update(formatted_resource)

    def force_resource_update(self, id):
        self.cvx_ids.discard(id)

    def _delete_neutron_resource(self, id):
        # Until we start using etcd, we need to unconditionally send the
        # delete request because it might have been created by another worker.
        # We force this by adding the resource to our 'view' of cvx resources
        self.cvx_ids.add(id)
        try:
            del self.neutron_resources[id]
            LOG.debug("%(pid)s %(class)s resource %(id)s removed locally",
                      {'class': self.__class__.__name__, 'id': id,
                       'pid': os.getpid()})
        except KeyError:
            LOG.debug("Resource ID %(id)s already deleted locally", {'id': id})

    def get_endpoint(self):
        return self.endpoint % {'region': self.region}

    @classmethod
    def get_resource_ids(cls, resource):
        return set([resource[cls.id_key]])

    def get_cvx_ids(self):
        if self.cvx_data_stale:
            LOG.info("%(pid)s Getting %(class)s from CVX",
                     {'class': self.__class__.__name__,
                      'pid': os.getpid()})
            cvx_data = self.rpc.send_api_request(self.get_endpoint(), 'GET')
            for resource in cvx_data:
                self.cvx_ids |= self.get_resource_ids(resource)
            self.cvx_data_stale = False
        return self.cvx_ids

    @staticmethod
    def get_db_resources(key=None):
        raise NotImplementedError

    def get_neutron_ids(self):
        if self.neutron_data_stale:
            self.get_neutron_resources()
        return set(self.neutron_resources.keys())

    def get_neutron_resources(self):
        if self.neutron_data_stale:
            LOG.info("%(pid)s Getting %(class)s from neutron",
                     {'class': self.__class__.__name__,
                      'pid': os.getpid()})
            for resource in self.get_db_resources():
                self._add_neutron_resource(resource)
            self.neutron_data_stale = False
        return self.neutron_resources

    def resource_ids_to_delete(self):
        cvx_resource_ids = self.get_cvx_ids()
        neutron_resource_ids = self.get_neutron_ids()
        return (cvx_resource_ids - neutron_resource_ids)

    def resource_ids_to_create(self):
        cvx_resource_ids = self.get_cvx_ids()
        neutron_resource_ids = self.get_neutron_ids()
        return (neutron_resource_ids - cvx_resource_ids)

    @classmethod
    def format_for_create(cls, neutron_resource):
        cvx_resource = dict(
            attr.transform(neutron_resource) for attr in cls.formatter
        )
        return {cvx_resource[cls.id_key]: cvx_resource}

    @classmethod
    def format_for_delete(cls, id):
        return {cls.id_key: id}

    def create_cvx_resources(self):
        resource_ids_to_create = self.resource_ids_to_create()
        neutron_resources = self.get_neutron_resources()
        resources_to_create = list(neutron_resources[resource_id] for
                                   resource_id in resource_ids_to_create)
        if resources_to_create:
            LOG.info("%(pid)s Creating %(class)s resources with ids %(ids)s "
                     "on CVX",
                     {'class': self.__class__.__name__,
                      'ids': ', '.join(str(r) for r in resource_ids_to_create),
                      'pid': os.getpid()})
            self.rpc.send_api_request(self.get_endpoint(), 'POST',
                                      resources_to_create)
            self.cvx_ids.update(resource_ids_to_create)
            LOG.info("%(pid)s %(class)s resources with ids %(ids)s created "
                     "on CVX",
                     {'class': self.__class__.__name__,
                      'ids': ', '.join(str(r) for r in resource_ids_to_create),
                      'pid': os.getpid()})
        else:
            LOG.debug("%(pid)s No %(class)s resources to create",
                      {'class': self.__class__.__name__,
                       'pid': os.getpid()})
        return resources_to_create

    def delete_cvx_resources(self):
        resource_ids_to_delete = self.resource_ids_to_delete()
        resources_to_delete = list(self.format_for_delete(id) for id in
                                   resource_ids_to_delete)
        if resources_to_delete:
            LOG.info("%(pid)s Deleting %(class)s resources with ids %(ids)s "
                     "from CVX",
                     {'class': self.__class__.__name__,
                      'ids': ', '.join(str(r) for r in resource_ids_to_delete),
                      'pid': os.getpid()})
            try:
                self.rpc.send_api_request(self.get_endpoint(), 'DELETE',
                                          resources_to_delete)
            except arista_exc.AristaRpcError as err:
                if not err.msg.startswith('Unknown port id'):
                    raise
            self.cvx_ids -= resource_ids_to_delete
            LOG.info("%(pid)s %(class)s resources with ids %(ids)s deleted "
                     "from CVX",
                     {'class': self.__class__.__name__,
                      'ids': ', '.join(str(r) for r in resource_ids_to_delete),
                      'pid': os.getpid()})
        else:
            LOG.debug("%(pid)s No %(class)s resources to delete",
                      {'class': self.__class__.__name__,
                       'pid': os.getpid()})
        return resources_to_delete


class Tenants(AristaResourcesBase):

    endpoint = 'region/%(region)s/tenant'
    formatter = [AttributeFormatter('project_id', 'id')]
    get_db_resources = staticmethod(db_lib.get_tenants)


class Networks(AristaResourcesBase):

    def _is_shared(rbac_entries, resource):
        for entry in rbac_entries:
            if (entry.action == 'access_as_shared' and
                    entry.target_tenant == '*'):
                return True
        return False

    def modify_blank_project_id(project_id, resource):
        if len(project_id) == 0:
            l3ha_network = getattr(resource, 'L3HARouterNetwork')
            if l3ha_network:
                project_id = l3ha_network.project_id
        return project_id

    formatter = [AttributeFormatter('id', 'id',
                                    submodel='Network'),
                 AttributeFormatter('project_id', 'tenantId',
                                    modify_blank_project_id,
                                    submodel='Network'),
                 AttributeFormatter('name', 'name',
                                    submodel='Network'),
                 AttributeFormatter('rbac_entries', 'shared', _is_shared,
                                    submodel='Network')]
    endpoint = 'region/%(region)s/network'
    get_db_resources = staticmethod(db_lib.get_networks)


class Segments(AristaResourcesBase):

    formatter = [AttributeFormatter('id', 'id'),
                 AttributeFormatter('network_type', 'type'),
                 AttributeFormatter('segmentation_id', 'segmentationId'),
                 AttributeFormatter('network_id', 'networkId'),
                 AttributeFormatter('is_dynamic', 'segmentType',
                                    lambda n, r: 'dynamic' if n else 'static')]
    endpoint = 'region/%(region)s/segment'
    get_db_resources = staticmethod(db_lib.get_segments)


class Dhcps(AristaResourcesBase):

    formatter = [AttributeFormatter('device_id', 'id',
                                    submodel='Port'),
                 AttributeFormatter('host', 'hostId',
                                    utils.hostname,
                                    submodel='PortBinding'),
                 AttributeFormatter('project_id', 'tenantId',
                                    submodel='Port')]
    endpoint = 'region/%(region)s/dhcp'
    get_db_resources = staticmethod(db_lib.get_dhcp_instances)


class Routers(AristaResourcesBase):

    def modify_blank_project_id(project_id, resource):
        if len(project_id) == 0:
            l3ha_router = getattr(resource, 'Router')
            if l3ha_router:
                project_id = l3ha_router.project_id
        return project_id

    formatter = [AttributeFormatter('device_id', 'id',
                                    submodel='Port'),
                 AttributeFormatter('device_owner', 'hostId',
                                    lambda *args: '(see router ports)',
                                    submodel='Port'),
                 AttributeFormatter('project_id', 'tenantId',
                                    modify_blank_project_id,
                                    submodel='Port')]
    endpoint = 'region/%(region)s/router'
    get_db_resources = staticmethod(db_lib.get_router_instances)


class Vms(AristaResourcesBase):

    formatter = [AttributeFormatter('device_id', 'id',
                                    submodel='Port'),
                 AttributeFormatter('host', 'hostId',
                                    utils.hostname,
                                    submodel='PortBinding'),
                 AttributeFormatter('project_id', 'tenantId',
                                    submodel='Port')]
    endpoint = 'region/%(region)s/vm'
    get_db_resources = staticmethod(db_lib.get_vm_instances)


class Baremetals(AristaResourcesBase):

    formatter = [AttributeFormatter('device_id', 'id',
                                    submodel='Port'),
                 AttributeFormatter('host', 'hostId',
                                    submodel='PortBinding'),
                 AttributeFormatter('project_id', 'tenantId',
                                    submodel='Port')]
    endpoint = 'region/%(region)s/baremetal'
    get_db_resources = staticmethod(db_lib.get_baremetal_instances)


class DhcpPorts(AristaResourcesBase):

    endpoint = 'region/%(region)s/port?type=dhcp'
    formatter = [AttributeFormatter('id', 'id', submodel='Port'),
                 AttributeFormatter('name', 'portName', submodel='Port'),
                 AttributeFormatter('device_owner', 'vlanType',
                                    lambda *args: 'allowed',
                                    submodel='Port'),
                 AttributeFormatter('network_id', 'networkId',
                                    submodel='Port'),
                 AttributeFormatter('device_id', 'instanceId',
                                    submodel='Port'),
                 AttributeFormatter('device_owner', 'instanceType',
                                    lambda *args: 'dhcp',
                                    submodel='Port'),
                 AttributeFormatter('project_id', 'tenantId',
                                    submodel='Port')]
    get_db_resources = staticmethod(db_lib.get_dhcp_ports)


class RouterPorts(AristaResourcesBase):

    def modify_blank_project_id(project_id, resource):
        if len(project_id) == 0:
            l3ha_router = getattr(resource, 'Router')
            if l3ha_router:
                project_id = l3ha_router.project_id
        return project_id

    endpoint = 'region/%(region)s/port?type=router'
    formatter = [AttributeFormatter('id', 'id', submodel='Port'),
                 AttributeFormatter('name', 'portName', submodel='Port'),
                 AttributeFormatter('device_owner', 'vlanType',
                                    lambda *args: 'allowed',
                                    submodel='Port'),
                 AttributeFormatter('network_id', 'networkId',
                                    submodel='Port'),
                 AttributeFormatter('device_id', 'instanceId',
                                    submodel='Port'),
                 AttributeFormatter('device_owner', 'instanceType',
                                    lambda *args: 'router',
                                    submodel='Port'),
                 AttributeFormatter('project_id', 'tenantId',
                                    modify_blank_project_id, submodel='Port')]
    get_db_resources = staticmethod(db_lib.get_router_ports)


class VmPorts(AristaResourcesBase):

    endpoint = 'region/%(region)s/port?type=vm'
    formatter = [AttributeFormatter('id', 'id', submodel='Port'),
                 AttributeFormatter('name', 'portName', submodel='Port'),
                 AttributeFormatter('device_owner', 'vlanType',
                                    lambda *args: 'allowed', submodel='Port'),
                 AttributeFormatter('network_id', 'networkId',
                                    submodel='Port'),
                 AttributeFormatter('device_id', 'instanceId',
                                    submodel='Port'),
                 AttributeFormatter('device_owner', 'instanceType',
                                    lambda *args: 'vm', submodel='Port'),
                 AttributeFormatter('project_id', 'tenantId',
                                    submodel='Port')]
    get_db_resources = staticmethod(db_lib.get_vm_ports)

    @classmethod
    def format_for_create(cls, resource):
        # This is needed until we can update the upstream trunk port
        # handling to add device_id to subports
        port = getattr(resource, 'Port')
        if port['device_owner'] == t_const.TRUNK_SUBPORT_OWNER:
            parent_port = db_lib.get_parent(port['id'])
            port['device_id'] = parent_port.get('device_id')
        return super(VmPorts, cls).format_for_create(resource)


class BaremetalPorts(AristaResourcesBase):

    def _get_vlan_type(device_owner, resource):
        if (device_owner.startswith(n_const.DEVICE_OWNER_BAREMETAL_PREFIX) or
                device_owner.startswith(n_const.DEVICE_OWNER_COMPUTE_PREFIX)):
            return 'native'
        else:
            return 'allowed'

    endpoint = 'region/%(region)s/port?type=baremetal'
    formatter = [AttributeFormatter('id', 'id', submodel='Port'),
                 AttributeFormatter('name', 'portName', submodel='Port'),
                 AttributeFormatter('device_owner', 'vlanType',
                                    _get_vlan_type, submodel='Port'),
                 AttributeFormatter('network_id', 'networkId',
                                    submodel='Port'),
                 AttributeFormatter('device_id', 'instanceId',
                                    submodel='Port'),
                 AttributeFormatter('device_owner', 'instanceType',
                                    lambda *args: 'baremetal',
                                    submodel='Port'),
                 AttributeFormatter('project_id', 'tenantId', submodel='Port')]
    get_db_resources = staticmethod(db_lib.get_baremetal_ports)


class PortBindings(AristaResourcesBase):

    endpoint = 'region/%(region)s/portbinding'
    get_db_resources = staticmethod(db_lib.get_port_bindings)

    @staticmethod
    def maybe_strip_fqdn(id):
        port_id, binding = id
        if type(binding) != tuple:
            binding = utils.hostname(binding)
        return (port_id, binding)

    def force_resource_update(self, id):
        id = self.maybe_strip_fqdn(id)
        super(PortBindings, self).force_resource_update(id)

    def _delete_neutron_resource(self, id):
        id = self.maybe_strip_fqdn(id)
        super(PortBindings, self)._delete_neutron_resource(id)

    @classmethod
    def get_resource_ids(cls, resource):
        resource_ids = set()
        port_id = resource['portId']
        for host_binding in resource.get('hostBinding', []):
            resource_ids.add((port_id, host_binding['host']))
        for switch_binding in resource.get('switchBinding', []):
            resource_ids.add((port_id, (switch_binding['switch'],
                                        switch_binding['interface'])))
        return resource_ids

    @classmethod
    def format_for_delete(cls, id):
        model = dict()
        port_id, binding = id
        model['portId'] = port_id
        if type(binding) == tuple:
            switch, interface = binding
            model['switchBinding'] = [{'switch': switch,
                                       'interface': interface}]
        else:
            host = binding
            model['hostBinding'] = [{'host': host}]
        return model

    @classmethod
    def format_for_create(cls, binding):
        cvx_resources = {}
        # First build the list of segments to which the port is bound
        # binding levels are in order from       0 -> highest
        #               which is typically   vxlan -> vlan
        # The Arista JSON API depends on this ordering being present in
        # the segments list
        segments = []
        for binding_level in (sorted(binding['levels'],
                                     key=lambda bl: bl.level)):
            segments.append({'id': binding_level.segment_id})

        # Determine if this is a switch or host bindings and populate
        # the appropriate model attribute accordingly
        host = utils.hostname(binding['host'])
        port_id = binding['port_id']
        # If the binding profile isn't valid json, this is a host binding
        try:
            profile = json.loads(binding.profile)
        except ValueError:
            profile = {}
        if profile.get('local_link_information'):
            for link in profile['local_link_information']:
                switch_binding = {'host': host,
                                  'switch': link['switch_id'],
                                  'interface': link['port_id'],
                                  'segment': segments}
                binding_key = (link['switch_id'], link['port_id'])
                cvx_resources[(port_id, binding_key)] = {
                    'portId': port_id,
                    'hostBinding': [],
                    'switchBinding': [switch_binding]}
        else:
            cvx_resources[(port_id, host)] = {
                'portId': port_id,
                'hostBinding': [{'host': host, 'segment': segments}],
                'switchBinding': []}
        return cvx_resources



# Copyright 2014 Arista Networks, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import socket
import struct

from neutron.plugins.ml2.driver_context import NetworkContext  # noqa
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as const
from neutron_lib import context as nctx
from oslo_config import cfg
from oslo_log import log as logging

from networking_arista._i18n import _, _LI
from networking_arista.common import api
from networking_arista.common import db_lib
from networking_arista.common import exceptions as arista_exc

LOG = logging.getLogger(__name__)
cfg.CONF.import_group('l3_arista', 'networking_arista.common.config')

EOS_UNREACHABLE_MSG = _('Unable to reach EOS')
DEFAULT_VLAN = 1
MLAG_SWITCHES = 2
VIRTUAL_ROUTER_MAC = '00:11:22:33:44:55'
IPV4_BITS = 32
IPV6_BITS = 128

# This string-format-at-a-distance confuses pylint :(
# pylint: disable=too-many-format-args
router_in_vrf_v1 = {
    'router': {'create': ['vrf definition {0}',
                          'rd {1}',
                          'exit',
                          'ip routing vrf {0}'],
               'delete': ['no vrf definition {0}']},
    'interface': {'add': ['vlan {0}',
                          'exit',
                          'interface vlan {0}',
                          'vrf forwarding {1}',
                          'ip address {2}'],
                  'remove': ['no interface vlan {0}']}}

router_in_vrf_v2 = {
    'router': {'create': ['vrf instance {0}',
                          'rd {1}',
                          'exit',
                          'ip routing vrf {0}'],
               'delete': ['no vrf instance {0}']},
    'interface': {'add': ['vlan {0}',
                          'exit',
                          'interface vlan {0}',
                          'vrf {1}',
                          'ip address {2}'],
                  'remove': ['no interface vlan {0}']}}

router_in_default_vrf = {
    'router': {'create': [],   # Place holder for now.
               'delete': []},  # Place holder for now.

    'interface': {'add': ['ip routing',
                          'vlan {0}',
                          'exit',
                          'interface vlan {0}',
                          'ip address {2}'],
                  'remove': ['no interface vlan {0}']}}

router_in_default_vrf_v6 = {
    'router': {'create': [],
               'delete': []},

    'interface': {'add': ['ipv6 unicast-routing',
                          'vlan {0}',
                          'exit',
                          'interface vlan {0}',
                          'ipv6 enable',
                          'ipv6 address {2}'],
                  'remove': ['no interface vlan {0}']}}

additional_cmds_for_mlag = {
    'router': {'create': ['ip virtual-router mac-address {0}'],
               'delete': []},

    'interface': {'add': ['ip virtual-router address {0}'],
                  'remove': []}}

additional_cmds_for_mlag_v6 = {
    'router': {'create': [],
               'delete': []},

    'interface': {'add': ['ipv6 virtual-router address {0}'],
                  'remove': []}}

additional_cmds_for_default_route = {
    'add': ['ip route vrf {0} 0.0.0.0/0 {1}'],
    # Remove is used when updating a network to remove default gateway
    # when deleting an interface we don't need to delete the route
    'remove': ['no ip route vrf {0} 0.0.0.0/0']}


class AristaL3Driver(object):
    """Wraps Arista JSON RPC.

    All communications between Neutron and EOS are over JSON RPC.
    EOS - operating system used on Arista hardware
    Command API - JSON RPC API provided by Arista EOS
    """
    def __init__(self):
        self._servers = []
        self._hosts = []
        self._interfaceDict = None
        self._validate_config()
        host = cfg.CONF.l3_arista.primary_l3_host
        self._hosts.append(host)
        self._servers.append(self._make_eapi_client(host))
        self._mlag_configured = cfg.CONF.l3_arista.mlag_config
        self._vrf_syntax_v2_supported = None
        self._router_in_vrf = router_in_vrf_v2
        self._use_vrf = cfg.CONF.l3_arista.use_vrf
        self._vrf_default_route = False
        if self._use_vrf:
            self._vrf_default_route = cfg.CONF.l3_arista.vrf_default_route
            if self._vrf_default_route:
                # only subscribe for events if vrf default route is enabled
                self.subscribe()
        if self._mlag_configured:
            host = cfg.CONF.l3_arista.secondary_l3_host
            self._hosts.append(host)
            self._servers.append(self._make_eapi_client(host))
            self._additionalRouterCmdsDict = additional_cmds_for_mlag['router']
            self._additionalInterfaceCmdsDict = (
                additional_cmds_for_mlag['interface'])
        if self._use_vrf:
            self.routerDict = self._router_in_vrf['router']
            self._update_vrf_commands(keep_alive=False)
            self._interfaceDict = self._router_in_vrf['interface']
        else:
            self.routerDict = router_in_default_vrf['router']
            self._interfaceDict = router_in_default_vrf['interface']
        self._enable_cleanup = cfg.CONF.l3_arista.enable_cleanup
        self._protected_vlans = self._parse_protected_vlans(
            cfg.CONF.l3_arista.protected_vlans)

    def subscribe(self):
        # Subscribe to the events related to networks and subnets
        registry.subscribe(self.update_subnet, resources.SUBNET,
                           events.AFTER_UPDATE)
        registry.subscribe(self.update_network, resources.NETWORK,
                           events.AFTER_UPDATE)

    def update_subnet(self, resource, event, trigger, **kwargs):
        subnet_info = kwargs['subnet']

        if subnet_info['ip_version'] == 6:
            LOG.info('IPv6 networks not supported with L3 plugin')
            return
        ctx = nctx.get_admin_context()
        ml2_db = NetworkContext(self, ctx, {'id': subnet_info['network_id']})
        seg_id = ml2_db.network_segments[0]['segmentation_id']
        router_info = db_lib.get_subnet_gateway_ipv4(subnet_info['id'])
        if router_info:
            router_info['seg_id'] = seg_id
            router_name = self._arista_router_name(router_info['id'],
                                                   router_info['name'])
            self._delete_default_gateway(router_name)
            self.add_router_interface(ctx, router_info)
            self._setup_default_gateway(router_info)

    def _reset_network_default_route(self, network_id):
        router_info = self._prepare_network_default_gateway(network_id)
        if router_info:
            router_name = self._arista_router_name(router_info['id'],
                                                   router_info['name'])
            self._delete_default_gateway(router_name)
            self._add_network_default_gateway(network_id)

    def update_network(self, resource, event, trigger, **kwargs):
        network_info = kwargs['network']
        self._reset_network_default_route(network_info['id'])

    def _prepare_network_default_gateway(self, network_id):
        router_info = db_lib.get_network_gateway_ipv4(network_id)
        if not router_info:
            return

        ip_version = router_info['ip_version']
        if ip_version == 6:
            LOG.info('IPv6 networks not supported with L3 plugin')
            return

        ctx = nctx.get_admin_context()
        ml2_db = NetworkContext(self, ctx, {'id': network_id})
        seg_id = ml2_db.network_segments[0]['segmentation_id']
        router_info['seg_id'] = seg_id
        return router_info

    def _add_network_default_gateway(self, network_id):
        router_info = self._prepare_network_default_gateway(network_id)
        if router_info:
            ctx = nctx.get_admin_context()
            self.add_router_interface(ctx, router_info)
            self._setup_default_gateway(router_info)

    @staticmethod
    def _raise_invalid_protected_vlans(vlan_string):
        msg = '%s is not a valid vlan or vlan range' % vlan_string
        LOG.error(msg)
        raise arista_exc.AristaServicePluginConfigError(msg=msg)

    def _parse_protected_vlans(self, vlan_strings):
        # VLAN 1 is always protected as it exists by default on EOS
        vlans = set([1])
        for vlan_string in vlan_strings:
            vlan_parsed = vlan_string.split(':', 2)

            if len(vlan_parsed) > 2:
                self._raise_invalid_protected_vlans(vlan_string)

            try:
                min_vlan = int(vlan_parsed[0])
            except ValueError:
                self._raise_invalid_protected_vlans(vlan_string)

            try:
                max_vlan = int(vlan_parsed[-1])
            except ValueError:
                self._raise_invalid_protected_vlans(vlan_string)

            if not (const.MIN_VLAN_TAG <= min_vlan <=
                    max_vlan <= const.MAX_VLAN_TAG):
                self._raise_invalid_protected_vlans(vlan_string)
            vlans.update(range(min_vlan, max_vlan + 1))
        return vlans

    @staticmethod
    def _make_eapi_client(host):
        return api.EAPIClient(
            host,
            username=cfg.CONF.l3_arista.primary_l3_host_username,
            password=cfg.CONF.l3_arista.primary_l3_host_password,
            verify=False,
            timeout=cfg.CONF.l3_arista.conn_timeout
        )

    def _supports_vrf_instance(self, version):
        ver_tokens = version.split('.')
        if len(ver_tokens) < 2 or int(ver_tokens[0]) < 4:
            return False
        if int(ver_tokens[0]) == 4 and int(ver_tokens[1]) < 22:
            return False
        return True

    def _check_vrf_syntax_v2_support(self, host, keep_alive=True):
        cmds = ['show version']
        result = None
        try:
            result = self._run_eos_cmds(cmds, host, log_exception=False,
                                        keep_alive=keep_alive,
                                        update_vrf_commands=False)
            LOG.info('show version result %s' % result)
        except Exception:
            # We don't know what exception we got return None
            # At this moment we don't know what command we support for vrf
            # creation
            return None

        return result and self._supports_vrf_instance(result[0].get('version',
                                                                    ''))

    def _update_vrf_commands(self, keep_alive=True):
        # This assumes all switches run the same version. This needs to be
        # updated if we'll support distributed routing
        new_vrf_support = self._check_vrf_syntax_v2_support(
            self._servers[0], keep_alive=keep_alive)

        if new_vrf_support == self._vrf_syntax_v2_supported:
            return

        LOG.info(_LI('Updating VRF command supported: %s'),
                 'vrf instance' if new_vrf_support else 'vrf definition')
        self._vrf_syntax_v2_supported = new_vrf_support
        if self._vrf_syntax_v2_supported is False:
            self._router_in_vrf = router_in_vrf_v1
        else:
            self._router_in_vrf = router_in_vrf_v2
        # we don't need to update self.interfaceDict as it is updated by
        # _select_dicts function before it is used
        self.routerDict = self._router_in_vrf['router']

    def _validate_config(self):
        if cfg.CONF.l3_arista.get('primary_l3_host') == '':
            msg = _('Required option primary_l3_host is not set')
            LOG.error(msg)
            raise arista_exc.AristaServicePluginConfigError(msg=msg)
        if cfg.CONF.l3_arista.get('mlag_config'):
            if cfg.CONF.l3_arista.get('secondary_l3_host') == '':
                msg = _('Required option secondary_l3_host is not set')
                LOG.error(msg)
                raise arista_exc.AristaServicePluginConfigError(msg=msg)
        if cfg.CONF.l3_arista.get('primary_l3_host_username') == '':
            msg = _('Required option primary_l3_host_username is not set')
            LOG.error(msg)
            raise arista_exc.AristaServicePluginConfigError(msg=msg)

    def create_router_on_eos(self, router_name, rdm, server):
        """Creates a router on Arista HW Device.

        :param router_name: globally unique identifier for router/VRF
        :param rdm: A value generated by hashing router name
        :param server: Server endpoint on the Arista switch to be configured
        """
        cmds = []
        rd = "%s:%s" % (rdm, rdm)

        for c in self.routerDict['create']:
            cmds.append(c.format(router_name, rd))

        if self._mlag_configured:
            mac = VIRTUAL_ROUTER_MAC
            for c in self._additionalRouterCmdsDict['create']:
                cmds.append(c.format(mac))

        self._run_config_cmds(cmds, server)

    def delete_router_from_eos(self, router_name, server):
        """Deletes a router from Arista HW Device.

        :param router_name: globally unique identifier for router/VRF
        :param server: Server endpoint on the Arista switch to be configured
        """
        cmds = []
        for c in self.routerDict['delete']:
            cmds.append(c.format(router_name))
        if self._mlag_configured:
            for c in self._additionalRouterCmdsDict['delete']:
                cmds.append(c)

        self._run_config_cmds(cmds, server)

    def _select_dicts(self, ipv):
        if self._use_vrf:
            if ipv == 6:
                msg = (_('IPv6 subnets are not supported with VRFs'))
                LOG.info(msg)

            self._interfaceDict = self._router_in_vrf['interface']
        else:
            if ipv == 6:
                # for IPv6 use IPv6 commands
                self._interfaceDict = router_in_default_vrf_v6['interface']
                self._additionalInterfaceCmdsDict = (
                    additional_cmds_for_mlag_v6['interface'])
            else:
                self._interfaceDict = router_in_default_vrf['interface']
                self._additionalInterfaceCmdsDict = (
                    additional_cmds_for_mlag['interface'])

    def add_interface_to_router(self, segment_id,
                                router_name, fixed_ip, router_ip, mask,
                                server):
        """Adds an interface to existing HW router on Arista HW device.

        :param segment_id: VLAN Id associated with interface that is added
        :param router_name: globally unique identifier for router/VRF
        :param fixed_ip: Fixed IP associated with the port
        :param router_ip: IP address of the router
        :param mask: subnet mask to be used
        :param server: Server endpoint on the Arista switch to be configured
        """

        if not segment_id:
            segment_id = DEFAULT_VLAN
        cmds = []
        for c in self._interfaceDict['add']:
            if self._mlag_configured:
                # In VARP config, use router ID else, use fixed IP.
                # If fixed Ip was not set this will be gateway IP address.
                ip = router_ip
            else:
                ip = fixed_ip + '/' + mask
            cmds.append(c.format(segment_id, router_name, ip))
        if self._mlag_configured:
            for c in self._additionalInterfaceCmdsDict['add']:
                cmds.append(c.format(fixed_ip))

        self._run_config_cmds(cmds, server)

    def delete_interface_from_router(self, segment_id, router_name, server):
        """Deletes an interface from existing HW router on Arista HW device.

        :param segment_id: VLAN Id associated with interface that is added
        :param router_name: globally unique identifier for router/VRF
        :param server: Server endpoint on the Arista switch to be configured
        """

        if not segment_id:
            segment_id = DEFAULT_VLAN
        cmds = []
        for c in self._interfaceDict['remove']:
            cmds.append(c.format(segment_id))

        self._run_config_cmds(cmds, server)

    def create_router(self, context, router):
        """Creates a router on Arista Switch.

        Deals with multiple configurations - such as Router per VRF,
        a router in default VRF, Virtual Router in MLAG configurations
        """
        if router:
            router_name = self._arista_router_name(router['id'],
                                                   router['name'])

            hashed = hashlib.sha256(router_name.encode('utf-8'))
            rdm = str(int(hashed.hexdigest(), 16) % 65536)

            mlag_peer_failed = False
            for s in self._servers:
                try:
                    self.create_router_on_eos(router_name, rdm, s)
                    mlag_peer_failed = False
                except Exception:
                    if self._mlag_configured and not mlag_peer_failed:
                        # In paired switch, it is OK to fail on one switch
                        mlag_peer_failed = True
                    else:
                        msg = (_('Failed to create router %s on EOS') %
                               router_name)
                        LOG.exception(msg)
                        raise arista_exc.AristaServicePluginRpcError(msg=msg)

            if self._vrf_default_route:
                ext_gateway = router.get('external_gateway_info')
                if ext_gateway:
                    network_id = ext_gateway.get('network_id')
                    if network_id:
                        self._add_network_default_gateway(network_id)

    def delete_router(self, context, router_id, router):
        """Deletes a router from Arista Switch."""

        if router:
            router_name = self._arista_router_name(router_id, router['name'])
            mlag_peer_failed = False
            for s in self._servers:
                try:
                    self.delete_router_from_eos(router_name, s)
                    mlag_peer_failed = False
                except Exception:
                    if self._mlag_configured and not mlag_peer_failed:
                        # In paired switch, it is OK to fail on one switch
                        mlag_peer_failed = True
                    else:
                        msg = (_('Failed to create router %s on EOS') %
                               router_name)
                        LOG.exception(msg)
                        raise arista_exc.AristaServicePluginRpcError(msg=msg)

    def update_router(self, context, router_id, original_router, new_router):
        """Updates a router which is already created on Arista Switch.

        """
        if not self._vrf_default_route or not new_router:
            return

        ext_gateway = new_router.get('external_gateway_info')
        if ext_gateway is None:
            # Remove default gateway if it exists
            orig_ext_gateway = original_router.get('external_gateway_info')
            if orig_ext_gateway is None:
                # External gateway did not change
                return
            network_id = orig_ext_gateway['network_id']
            ml2_db = NetworkContext(self, context, {'id': network_id})
            seg_id = ml2_db.network_segments[0]['segmentation_id']
            new_router['seg_id'] = seg_id
            new_router['ip_version'] = 4
            self.remove_router_interface(context, new_router,
                                         delete_gateway=True)
            return

        network_id = ext_gateway.get('network_id')
        if network_id:
            self._reset_network_default_route(network_id)

    def _setup_default_gateway(self, router_info):
        mlag_peer_failed = False
        gip = router_info['gip']
        router_name = self._arista_router_name(router_info['id'],
                                               router_info['name'])

        for s in self._servers:
            try:
                self._setup_switch_default_gateway(router_name, gip, s)
                mlag_peer_failed = False
            except Exception:
                if self._mlag_configured and not mlag_peer_failed:
                    # In paired switch, it is OK to fail on one switch
                    mlag_peer_failed = True
                else:
                    msg = (_('Failed to setup router gateway %s on EOS') %
                           router_name)
                    LOG.exception(msg)
                    raise arista_exc.AristaServicePluginRpcError(msg=msg)

    def _setup_switch_default_gateway(self, router_name, gip, server):
        cmds = [
            c.format(router_name, gip)
            for c in additional_cmds_for_default_route['add']
        ]

        self._run_config_cmds(cmds, server)

    def _delete_default_gateway(self, router_name):
        mlag_peer_failed = False

        for s in self._servers:
            try:
                self._delete_switch_default_gateway(router_name, s)
                mlag_peer_failed = False
            except Exception:
                if self._mlag_configured and not mlag_peer_failed:
                    # In paired switch, it is OK to fail on one switch
                    mlag_peer_failed = True
                else:
                    msg = (_('Failed to delete router gateway %s on EOS') %
                           router_name)
                    LOG.exception(msg)
                    raise arista_exc.AristaServicePluginRpcError(msg=msg)

    def _delete_switch_default_gateway(self, router_name, server):
        cmds = [
            c.format(router_name)
            for c in additional_cmds_for_default_route['remove']
        ]

        self._run_config_cmds(cmds, server)

    def add_router_interface(self, context, router_info):
        """Adds an interface to a router created on Arista HW router.

        This deals with both IPv6 and IPv4 configurations.
        """
        if router_info:
            if router_info['ip_version'] == 6 and self._use_vrf:
                # For the moment we ignore the interfaces to be added
                # on IPv6 subnets.
                LOG.info('Using VRFs. Ignoring IPv6 interface')
                return
            self._select_dicts(router_info['ip_version'])
            cidr = router_info['cidr']
            subnet_mask = cidr.split('/')[1]
            router_name = self._arista_router_name(router_info['id'],
                                                   router_info['name'])
            if self._mlag_configured:
                # For MLAG, we send a specific IP address as opposed to cidr
                # For now, we are using x.x.x.253 and x.x.x.254 as virtual IP
                # unless either collides with the router interface fixed_ip,
                # in which case we use x.x.x.252
                mlag_peer_failed = False
                router_ips = self._get_router_ips(cidr, len(self._servers),
                                                  router_info['ip_version'],
                                                  router_info['fixed_ip'])
                for i, server in enumerate(self._servers):
                    # Get appropriate virtual IP address for this router
                    router_ip = router_ips[i]
                    try:
                        self.add_interface_to_router(router_info['seg_id'],
                                                     router_name,
                                                     router_info['fixed_ip'],
                                                     router_ip, subnet_mask,
                                                     server)
                        mlag_peer_failed = False
                    except Exception:
                        if not mlag_peer_failed:
                            mlag_peer_failed = True
                        else:
                            msg = (_('Failed to add interface to router '
                                     '%s on EOS') % router_name)
                            LOG.exception(msg)
                            raise arista_exc.AristaServicePluginRpcError(
                                msg=msg)

            else:
                for s in self._servers:
                    self.add_interface_to_router(router_info['seg_id'],
                                                 router_name,
                                                 router_info['fixed_ip'],
                                                 None, subnet_mask, s)

    def remove_router_interface(self, context, router_info,
                                delete_gateway=False):
        """Removes previously configured interface from router on Arista HW.

        This deals with both IPv6 and IPv4 configurations.
        """
        if router_info:
            if router_info['ip_version'] == 6 and self._use_vrf:
                # For the moment we ignore the interfaces to be added
                # on IPv6 subnets.
                LOG.info('Using VRFs. Ignoring IPv6 interface')
                return
            router_name = self._arista_router_name(router_info['id'],
                                                   router_info['name'])
            mlag_peer_failed = False
            for s in self._servers:
                try:
                    if delete_gateway:
                        self._delete_switch_default_gateway(router_name, s)
                    self.delete_interface_from_router(router_info['seg_id'],
                                                      router_name, s)
                    if self._mlag_configured:
                        mlag_peer_failed = False
                except Exception:
                    if self._mlag_configured and not mlag_peer_failed:
                        mlag_peer_failed = True
                    else:
                        msg = (_('Failed to remove interface to router '
                                 '%s on EOS') % router_name)
                        LOG.exception(msg)
                        raise arista_exc.AristaServicePluginRpcError(msg=msg)

    def _run_config_cmds(self, commands, server, log_exception=True,
                         keep_alive=True, update_vrf_commands=True):
        """Execute/sends a CAPI (Command API) command to EOS.

        In this method, list of commands is appended with prefix and
        postfix commands - to make is understandble by EOS.

        :param commands : List of command to be executed on EOS.
        :param server: Server endpoint on the Arista switch to be configured
        """
        command_start = ['enable', 'configure']
        command_end = ['exit']
        full_command = command_start + commands + command_end
        self._run_eos_cmds(full_command, server, log_exception, keep_alive,
                           update_vrf_commands)

    def _run_eos_cmds(self, commands, server, log_exception=True,
                      keep_alive=True, update_vrf_commands=True):
        LOG.info(_LI('Executing command on Arista EOS: %s'), commands)

        try:
            # this returns array of return values for every command in
            # full_command list
            ret = server.execute(commands, keep_alive=keep_alive)
            LOG.info(_LI('Results of execution on Arista EOS: %s'), ret)
            return ret
        except arista_exc.AristaServicePluginInvalidCommand:
            msg = (_('VRF creation command unsupported. This request should '
                     'work on next retry.'))
            if log_exception:
                LOG.exception(msg)
            if self._use_vrf and update_vrf_commands:
                # For now we assume that the only command that raises this
                # exception is vrf instance/definition and we need to update
                # the current support
                self._update_vrf_commands()
            raise
        except Exception:
            msg = (_('Error occurred while trying to execute '
                     'commands %(cmd)s on EOS %(host)s') %
                   {'cmd': commands, 'host': server})
            if log_exception:
                LOG.exception(msg)
            raise arista_exc.AristaServicePluginRpcError(msg=msg)

    def _arista_router_name(self, router_id, name):
        """Generate an arista specific name for this router.

        Use a unique name so that OpenStack created routers/SVIs
        can be distinguishged from the user created routers/SVIs
        on Arista HW. Replace spaces with underscores for CLI compatibility
        """
        return '__OpenStack__' + router_id + '-' + name.replace(' ', '_')

    def _get_binary_from_ipv4(self, ip_addr):
        """Converts IPv4 address to binary form."""

        return struct.unpack("!L", socket.inet_pton(socket.AF_INET,
                                                    ip_addr))[0]

    def _get_binary_from_ipv6(self, ip_addr):
        """Converts IPv6 address to binary form."""

        hi, lo = struct.unpack("!QQ", socket.inet_pton(socket.AF_INET6,
                                                       ip_addr))
        return (hi << 64) | lo

    def _get_ipv4_from_binary(self, bin_addr):
        """Converts binary address to Ipv4 format."""

        return socket.inet_ntop(socket.AF_INET, struct.pack("!L", bin_addr))

    def _get_ipv6_from_binary(self, bin_addr):
        """Converts binary address to Ipv6 format."""

        hi = bin_addr >> 64
        lo = bin_addr & 0xFFFFFFFF
        return socket.inet_ntop(socket.AF_INET6, struct.pack("!QQ", hi, lo))

    def _get_router_ips(self, cidr, ip_count, ip_ver, fixed_ip):
        """For a given IP subnet and IP version type, generate IP for router.

        This method takes the network address (cidr) and selects a set of
        IP addresses that should be assigned to virtual router running
        on multiple switches. It uses upper addresses in a subnet address
        as IP for the router. Each instace of the router, on each switch,
        requires uniqe IP address. For example in IPv4 case, on a 255
        subnet, it will pick X.X.X.254 as first addess, X.X.X.253 for next,
        and so on, skipping an address specified as the reserved_ip
        """
        network_addr, prefix = cidr.split('/')
        if ip_ver == 4:
            bits = IPV4_BITS
            ip = self._get_binary_from_ipv4(network_addr)
        elif ip_ver == 6:
            bits = IPV6_BITS
            ip = self._get_binary_from_ipv6(network_addr)

        mask = (pow(2, bits) - 1) << (bits - int(prefix))

        network_addr = ip & mask

        router_ips = list()
        ip_idx = 0
        while len(router_ips) < ip_count:
            start_ip = MLAG_SWITCHES + ip_idx
            router_ip = pow(2, bits - int(prefix)) - start_ip

            router_ip = network_addr | router_ip
            if ip_ver == 4:
                router_ip = self._get_ipv4_from_binary(router_ip)
            else:
                router_ip = self._get_ipv6_from_binary(router_ip)
            ip_idx += 1
            if router_ip != fixed_ip.lower():
                router_ips.append(router_ip + '/' + prefix)
        return router_ips



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import json
import re

from neutron_lib import constants as n_const
from oslo_config import cfg
from oslo_log import log as logging

from networking_arista._i18n import _LI
from networking_arista.common import api
from networking_arista.common import constants as a_const
from networking_arista.common import db_lib
from networking_arista.common import exceptions as arista_exc
from networking_arista.common import utils

LOG = logging.getLogger(__name__)


class AristaSecurityGroupSwitchHelper(object):
    """Helper class for applying baremetal security groups on Arista switches

    This helper class contains methods for adding and removing security
    groups, security group rules and security group port bindings to and from
    Arista switches.
    """

    def initialize_switch_endpoints(self):
        """Initialize endpoints for switch communication"""
        self._switches = {}
        self._port_group_info = {}
        self._validate_config()
        for s in cfg.CONF.ml2_arista.switch_info:
            switch_ip, switch_user, switch_pass = s.split(":")
            if switch_pass == "''":
                switch_pass = ''
            self._switches[switch_ip] = api.EAPIClient(
                switch_ip,
                switch_user,
                switch_pass,
                verify=False,
                timeout=cfg.CONF.ml2_arista.conn_timeout)
        self._check_dynamic_acl_support()

    def _check_dynamic_acl_support(self):
        """Log an error if any switches don't support dynamic ACLs"""
        cmds = ['ip access-list openstack-test dynamic',
                'no ip access-list openstack-test']
        for switch_ip, switch_client in self._switches.items():
            try:
                self.run_openstack_sg_cmds(cmds)
            except Exception:
                LOG.error("Switch %s does not support dynamic ACLs. SG "
                          "support will not be enabled on this switch.",
                          switch_ip)

    def _validate_config(self):
        """Ensure at least one switch is configured"""
        if len(cfg.CONF.ml2_arista.get('switch_info')) < 1:
            msg = _('Required option - when "sec_group_support" is enabled, '
                    'at least one switch must be specified ')
            LOG.exception(msg)
            raise arista_exc.AristaConfigError(msg=msg)

    def run_openstack_sg_cmds(self, commands, switch):
        """Execute/sends a CAPI (Command API) command to EOS.

        In this method, list of commands is appended with prefix and
        postfix commands - to make is understandble by EOS.

        :param commands : List of command to be executed on EOS.
        :param switch: Endpoint on the Arista switch to be configured
        """
        if not switch:
            LOG.exception("No client found for switch")
            return []
        if len(commands) == 0:
            return []
        command_start = ['enable', 'configure']
        command_end = ['exit']
        full_command = command_start + commands + command_end
        return self._run_eos_cmds(full_command, switch)

    def _run_eos_cmds(self, commands, switch):
        """Execute/sends a CAPI (Command API) command to EOS.

        This method is useful for running show commands that require no
        prefix or postfix commands.

        :param commands : List of commands to be executed on EOS.
        :param switch: Endpoint on the Arista switch to be configured
        """
        LOG.info(_LI('Executing command on Arista EOS: %s'), commands)

        try:
            # this returns array of return values for every command in
            # commands list
            ret = switch.execute(commands)
            LOG.info(_LI('Results of execution on Arista EOS: %s'), ret)
            return ret
        except Exception:
            msg = (_('Error occurred while trying to execute '
                     'commands %(cmd)s on EOS %(host)s') %
                   {'cmd': commands, 'host': switch})
            LOG.exception(msg)

    @staticmethod
    def _acl_name(name, direction):
        """Generate an arista specific name for this ACL.

        Use a unique name so that OpenStack created ACLs
        can be distinguishged from the user created ACLs
        on Arista HW.
        """
        direction = direction.upper()
        return 'SG' + '-' + direction + '-' + name

    @staticmethod
    def _get_switchports(profile):
        """Return list of (switch_ip, interface) tuples from local_link_info"""
        switchports = []
        if profile.get('local_link_information'):
            for link in profile['local_link_information']:
                if 'switch_info' in link and 'port_id' in link:
                    switch = link['switch_info']
                    interface = link['port_id']
                    switchports.append((switch, interface))
                else:
                    LOG.warning("Incomplete link information: %s", link)
        return switchports

    def _update_port_group_info(self, switches=None):
        """Refresh data on switch interfaces' port group membership"""
        if switches is None:
            switches = self._switches.keys()
        for switch_ip in switches:
            client = self._switches.get(switch_ip)
            ret = self._run_eos_cmds(['show interfaces'], client)
            if not ret or len(ret) == 0:
                LOG.warning("Unable to retrieve interface info for %s",
                            switch_ip)
                continue
            intf_info = ret[0]
            self._port_group_info[switch_ip] = intf_info.get('interfaces', {})

    def _get_port_for_acl(self, port_id, switch):
        """Gets interface name for ACLs

        Finds the Port-Channel name if port_id is in a Port-Channel, otherwise
        ACLs are applied to Ethernet interface.

        :param port_id: Name of port from ironic db
        :param server: Server endpoint on the Arista switch to be configured
        """
        all_intf_info = self._port_group_info.get(switch, {})
        intf_info = all_intf_info.get(port_id, {})
        member_info = intf_info.get('interfaceMembership', '')
        port_group_info = re.search('Member of (?P<port_group>\S+)',
                                    member_info)
        if port_group_info:
            port_id = port_group_info.group('port_group')
        return port_id

    @staticmethod
    def _supported_rule(protocol, ethertype):
        """Checks that the rule is an IPv4 rule of a supported protocol"""
        if not protocol or protocol not in utils.SUPPORTED_SG_PROTOCOLS:
            return False

        if ethertype != n_const.IPv4:
            return False

        return True

    def _format_rule(self, protocol, cidr, min_port, max_port, direction):
        """Get EOS formatted rule"""
        if cidr is None:
            cidr = 'any'

        if direction == n_const.INGRESS_DIRECTION:
            dst_ip = 'any'
            src_ip = cidr
        elif direction == n_const.EGRESS_DIRECTION:
            dst_ip = cidr
            src_ip = 'any'

        if protocol == n_const.PROTO_NAME_ICMP:
            rule = "permit icmp %s %s" % (src_ip, dst_ip)
            if min_port:
                rule += " %s" % (min_port)
                if max_port:
                    rule += " %s" % (max_port)
        else:
            rule = "permit %s %s %s" % (protocol, src_ip, dst_ip)
            if min_port and max_port:
                rule += " range %s %s" % (min_port, max_port)
            elif min_port and not max_port:
                rule += " eq %s" % min_port
        return rule

    def _format_rules_for_eos(self, rules):
        """Format list of rules for EOS and sort into ingress/egress rules"""
        in_rules = []
        eg_rules = []
        for rule in rules:
            protocol = rule.get('protocol')
            cidr = rule.get('remote_ip_prefix', 'any')
            min_port = rule.get('port_range_min')
            max_port = rule.get('port_range_max')
            direction = rule.get('direction')
            ethertype = rule.get('ethertype')
            if not self._supported_rule(protocol, ethertype):
                continue
            formatted_rule = self._format_rule(protocol, cidr, min_port,
                                               max_port, direction)
            if rule['direction'] == n_const.INGRESS_DIRECTION:
                in_rules.append(formatted_rule)
            elif rule['direction'] == n_const.EGRESS_DIRECTION:
                eg_rules.append(formatted_rule)
        return in_rules, eg_rules


class AristaSecurityGroupCallbackHelper(AristaSecurityGroupSwitchHelper):

    def run_cmds_on_all_switches(self, cmds):
        """Runs all cmds on all configured switches

        This helper is used for ACL and rule creation/deletion as ACLs
        and rules must exist on all switches.
        """
        for switch in self._switches.values():
            self.run_openstack_sg_cmds(cmds, switch)

    def run_per_switch_cmds(self, switch_cmds):
        """Applies cmds to appropriate switches

        This takes in a switch->cmds mapping and runs only the set of cmds
        specified for a switch on that switch. This helper is used for
        applying/removing ACLs to/from interfaces as this config will vary
        from switch to switch.
        """
        for switch_ip, cmds in switch_cmds.items():
            switch = self._switches.get(switch_ip)
            self.run_openstack_sg_cmds(cmds, switch)

    def _get_switches(self, profile):
        """Get set of switches referenced in a port binding profile"""
        switchports = self._get_switchports(profile)
        switches = set([switchport[0] for switchport in switchports])
        return switches

    def get_create_security_group_commands(self, sg_id, sg_rules):
        """Commands for creating ACL"""
        cmds = []
        in_rules, eg_rules = self._format_rules_for_eos(sg_rules)
        cmds.append("ip access-list %s dynamic" %
                    self._acl_name(sg_id, n_const.INGRESS_DIRECTION))
        for in_rule in in_rules:
            cmds.append(in_rule)
        cmds.append("exit")
        cmds.append("ip access-list %s dynamic" %
                    self._acl_name(sg_id, n_const.EGRESS_DIRECTION))
        for eg_rule in eg_rules:
            cmds.append(eg_rule)
        cmds.append("exit")
        return cmds

    def get_delete_security_group_commands(self, sg_id):
        """Commands for deleting ACL"""
        cmds = []
        cmds.append("no ip access-list %s" %
                    self._acl_name(sg_id, n_const.INGRESS_DIRECTION))
        cmds.append("no ip access-list %s" %
                    self._acl_name(sg_id, n_const.EGRESS_DIRECTION))
        return cmds

    def _get_rule_cmds(self, sg_id, sg_rule, delete=False):
        """Helper for getting add/delete ACL rule commands"""
        rule_prefix = ""
        if delete:
            rule_prefix = "no "
        in_rules, eg_rules = self._format_rules_for_eos([sg_rule])
        cmds = []
        if in_rules:
            cmds.append("ip access-list %s dynamic" %
                        self._acl_name(sg_id, n_const.INGRESS_DIRECTION))
            for in_rule in in_rules:
                cmds.append(rule_prefix + in_rule)
            cmds.append("exit")
        if eg_rules:
            cmds.append("ip access-list %s dynamic" %
                        self._acl_name(sg_id, n_const.EGRESS_DIRECTION))
            for eg_rule in eg_rules:
                cmds.append(rule_prefix + eg_rule)
            cmds.append("exit")
        return cmds

    def get_create_security_group_rule_commands(self, sg_id, sg_rule):
        """Commands for adding rule to ACL"""
        return self._get_rule_cmds(sg_id, sg_rule)

    def get_delete_security_group_rule_commands(self, sg_id, sg_rule):
        """Commands for removing rule from ACLS"""
        return self._get_rule_cmds(sg_id, sg_rule, delete=True)

    def _get_interface_commands(self, sg_id, profile, delete=False):
        """Helper for getting interface ACL apply/remove commands"""
        rule_prefix = ""
        if delete:
            rule_prefix = "no "
        switch_cmds = {}
        switchports = self._get_switchports(profile)
        for switch_ip, intf in switchports:
            cmds = []
            intf_id = self._get_port_for_acl(intf, switch_ip)
            cmds.append("interface %s" % intf_id)
            name = self._acl_name(sg_id, n_const.INGRESS_DIRECTION)
            cmds.append(rule_prefix + "ip access-group %s %s" %
                        (name, a_const.INGRESS_DIRECTION))
            name = self._acl_name(sg_id, n_const.EGRESS_DIRECTION)
            cmds.append(rule_prefix + "ip access-group %s %s" %
                        (name, a_const.EGRESS_DIRECTION))
            cmds.append("exit")
            if switch_ip not in switch_cmds.keys():
                switch_cmds[switch_ip] = []
            switch_cmds[switch_ip].extend(cmds)
        return switch_cmds

    def get_apply_security_group_commands(self, sg_id, profile):
        """Commands for applying ACL to interface"""
        return self._get_interface_commands(sg_id, profile)

    def get_remove_security_group_commands(self, sg_id, profile):
        """Commands for removing ACL from interface"""
        return self._get_interface_commands(sg_id, profile, delete=True)


class AristaSecurityGroupSyncHelper(AristaSecurityGroupSwitchHelper):

    def _parse_acl_config(self, acl_config):
        """Parse configured ACLs and rules

        ACLs are returned as a dict of rule sets:
        {<eos_acl1_name>: set([<eos_acl1_rules>]),
         <eos_acl2_name>: set([<eos_acl2_rules>]),
         ...,
        }
        """
        parsed_acls = dict()
        for acl in acl_config['aclList']:
            parsed_acls[acl['name']] = set()
            for rule in acl['sequence']:
                parsed_acls[acl['name']].add(rule['text'])
        return parsed_acls

    def _parse_binding_config(self, binding_config):
        """Parse configured interface -> ACL bindings

        Bindings are returned as a set of (intf, name, direction) tuples:
        set([(intf1, acl_name, direction),
             (intf2, acl_name, direction),
             ...,
        ])
        """
        parsed_bindings = set()
        for acl in binding_config['aclList']:
            for intf in acl['configuredIngressIntfs']:
                parsed_bindings.add((intf['name'], acl['name'],
                                     a_const.INGRESS_DIRECTION))
            for intf in acl['configuredEgressIntfs']:
                parsed_bindings.add((intf['name'], acl['name'],
                                     a_const.EGRESS_DIRECTION))
        return parsed_bindings

    def _get_dynamic_acl_info(self, switch_ip):
        """Retrieve ACLs, ACLs rules and interface bindings from switch"""
        cmds = ["enable",
                "show ip access-lists dynamic",
                "show ip access-lists summary dynamic"]
        switch = self._switches.get(switch_ip)

        _, acls, bindings = self._run_eos_cmds(cmds, switch)

        parsed_acls = self._parse_acl_config(acls)
        parsed_bindings = self._parse_binding_config(bindings)

        return parsed_acls, parsed_bindings

    def get_expected_acls(self):
        """Query the neutron DB for Security Groups and Rules

        Groups and rules are returned as a dict of rule sets:
        {<eos_acl1_name>: set([<eos_acl1_rules>]),
         <eos_acl2_name>: set([<eos_acl2_rules>]),
         ...,
        }
        """
        security_groups = db_lib.get_security_groups()

        expected_acls = collections.defaultdict(set)
        for sg in security_groups:
            in_rules, out_rules = self._format_rules_for_eos(sg['rules'])
            ingress_acl_name = self._acl_name(sg['id'],
                                              n_const.INGRESS_DIRECTION)
            egress_acl_name = self._acl_name(sg['id'],
                                             n_const.EGRESS_DIRECTION)
            expected_acls[ingress_acl_name].update(in_rules)
            expected_acls[egress_acl_name].update(out_rules)
        return expected_acls

    def get_expected_bindings(self):
        """Query the neutron DB for SG->switch interface bindings

        Bindings are returned as a dict of bindings for each switch:
        {<switch1>: set([(intf1, acl_name, direction),
                         (intf2, acl_name, direction)]),
         <switch2>: set([(intf1, acl_name, direction)]),
         ...,
        }
        """
        sg_bindings = db_lib.get_baremetal_sg_bindings()

        all_expected_bindings = collections.defaultdict(set)
        for sg_binding, port_binding in sg_bindings:
            sg_id = sg_binding['security_group_id']
            try:
                binding_profile = json.loads(port_binding.profile)
            except ValueError:
                binding_profile = {}
            switchports = self._get_switchports(binding_profile)
            for switch, intf in switchports:
                ingress_name = self._acl_name(sg_id, n_const.INGRESS_DIRECTION)
                egress_name = self._acl_name(sg_id, n_const.EGRESS_DIRECTION)
                all_expected_bindings[switch].add(
                    (intf, ingress_name, a_const.INGRESS_DIRECTION))
                all_expected_bindings[switch].add(
                    (intf, egress_name, a_const.EGRESS_DIRECTION))
        return all_expected_bindings

    def adjust_bindings_for_lag(self, switch_ip, bindings):
        """Adjusting interface names for expected bindings where LAGs exist"""

        # Get latest LAG info for switch
        self._update_port_group_info([switch_ip])

        # Update bindings to account for LAG info
        adjusted_bindings = set()
        for binding in bindings:
            adjusted_bindings.add(
                (self._get_port_for_acl(binding[0], switch_ip),) + binding[1:])
        return adjusted_bindings

    def get_sync_acl_cmds(self, switch_acls, expected_acls):
        """Returns the list of commands required synchronize switch ACLs

        1. Identify unexpected ACLs and delete them
        2. Iterate over expected ACLs
           a. Add missing ACLs + all rules
           b. Delete unexpected rules
           c. Add missing rules
        """
        switch_cmds = list()

        # Delete any stale ACLs
        acls_to_delete = (set(switch_acls.keys()) - set(expected_acls.keys()))
        for acl in acls_to_delete:
            switch_cmds.append('no ip access-list %s' % acl)

        # Update or create ACLs and rules
        for acl, expected_rules in expected_acls.items():
            switch_rules = switch_acls.get(acl, set())
            rules_to_delete = switch_rules - expected_rules
            rules_to_add = expected_rules - switch_rules
            # Check if ACL requires create or rule changes
            if (acl in switch_acls and
                    len(rules_to_add | rules_to_delete) == 0):
                continue
            switch_cmds.append('ip access-list %s dynamic' % acl)
            # Delete any stale rules
            for rule in rules_to_delete:
                switch_cmds.append('no ' + rule)
            # Add any missing rules
            for rule in rules_to_add:
                switch_cmds.append(rule)
                switch_cmds.append('exit')
        return switch_cmds

    def get_sync_binding_cmds(self, switch_bindings, expected_bindings):
        """Returns the list of commands required to synchronize ACL bindings

        1. Delete any unexpected bindings
        2. Add any missing bindings
        """
        switch_cmds = list()

        # Update any necessary switch interface ACLs
        bindings_to_delete = switch_bindings - expected_bindings
        bindings_to_add = expected_bindings - switch_bindings
        for intf, acl, direction in bindings_to_delete:
            switch_cmds.extend(['interface %s' % intf,
                                'no ip access-group %s %s' %
                                (acl, direction),
                                'exit'])
        for intf, acl, direction in bindings_to_add:
            switch_cmds.extend(['interface %s' % intf,
                                'ip access-group %s %s' % (acl, direction),
                                'exit'])
        return switch_cmds



# Copyright 2014 Arista Networks, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import traceback

from neutron_lib.agent import topics
from neutron_lib import constants as n_const
from neutron_lib import context as nctx
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from neutron_lib.services import base as service_base
from neutron_lib import worker
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import excutils

from neutron.api.rpc.agentnotifiers import l3_rpc_agent_api
from neutron.api.rpc.handlers import l3_rpc
from neutron.db import extraroute_db
from neutron.db import l3_agentschedulers_db
from neutron.db import l3_gwmode_db
from neutron.plugins.ml2.driver_context import NetworkContext  # noqa

from networking_arista._i18n import _LE, _LI
from networking_arista.l3Plugin import arista_l3_driver

LOG = logging.getLogger(__name__)


class AristaL3SyncWorker(worker.BaseWorker):
    def __init__(self, driver):
        self.driver = driver
        self._enable_cleanup = driver._enable_cleanup
        self._protected_vlans = driver._protected_vlans
        self._servers = driver._servers
        self._use_vrf = driver._use_vrf
        self._vrf_default_route = driver._vrf_default_route
        self._loop = None
        super(AristaL3SyncWorker, self).__init__(worker_process_count=0)

    def start(self):
        super(AristaL3SyncWorker, self).start()
        if self._loop is None:
            self._loop = loopingcall.FixedIntervalLoopingCall(
                self.synchronize
            )
        self._loop.start(interval=cfg.CONF.l3_arista.l3_sync_interval)

    def stop(self):
        if self._loop is not None:
            self._loop.stop()

    def wait(self):
        if self._loop is not None:
            self._loop.wait()
        self._loop = None

    def reset(self):
        self.stop()
        self.wait()
        self.start()

    def get_subnet_info(self, subnet_id):
        return self.get_subnet(subnet_id)

    def get_routers_and_interfaces(self):
        core = directory.get_plugin()
        ctx = nctx.get_admin_context()
        routers = directory.get_plugin(plugin_constants.L3).get_routers(ctx)
        router_interfaces = list()
        for r in routers:
            ports = core.get_ports(
                ctx,
                filters={
                    'device_id': [r['id']],
                    'device_owner': [n_const.DEVICE_OWNER_ROUTER_INTF,
                                     n_const.DEVICE_OWNER_ROUTER_GW]}) or []
            for p in ports:
                router_interface = r.copy()
                net_id = p['network_id']
                subnet_id = p['fixed_ips'][0]['subnet_id']
                subnet = core.get_subnet(ctx, subnet_id)
                ml2_db = NetworkContext(self, ctx, {'id': net_id})
                seg_id = ml2_db.network_segments[0]['segmentation_id']

                router_interface['seg_id'] = seg_id
                router_interface['cidr'] = subnet['cidr']
                router_interface['gip'] = subnet['gateway_ip']
                if p['id'] == r['gw_port_id']:
                    r['gip'] = subnet['gateway_ip']
                router_interface['fixed_ip'] = p['fixed_ips'][0]['ip_address']
                router_interface['ip_version'] = subnet['ip_version']
                router_interface['subnet_id'] = subnet_id
                router_interfaces.append(router_interface)
        return routers, router_interfaces

    def _get_ext_gateway_ip(self, router):
        if ('external_gateway_info' not in router or
                router['external_gateway_info'] is None):
            return None
        return router.get('gip')

    def synchronize(self):
        """Synchronizes Router DB from Neturon DB with EOS.

        Walks through the Neturon Db and ensures that all the routers
        created in Netuton DB match with EOS. After creating appropriate
        routers, it ensures to add interfaces as well.
        Uses idempotent properties of EOS configuration, which means
        same commands can be repeated.
        """
        LOG.info(_LI('Syncing Neutron Router DB <-> EOS'))
        try:
            if self._use_vrf:
                self.driver._update_vrf_commands()

            routers, router_interfaces = self.get_routers_and_interfaces()
            if self._enable_cleanup:
                expected_vrfs = set()
                expected_vrf_default_routes = set()
                if self._use_vrf:
                    expected_vrfs.update(self.driver._arista_router_name(
                        r['id'], r['name']) for r in routers)
                    if self._vrf_default_route:
                        expected_vrf_default_routes.update(
                            (self.driver._arista_router_name(router['id'],
                                                             router['name']),
                             self._get_ext_gateway_ip(router))
                            for router in routers)

                expected_vlans = set(r['seg_id'] for r in router_interfaces)
                LOG.info(_LI('Syncing Neutron Router DB - cleanup'))
                self.do_cleanup(expected_vrfs, expected_vlans,
                                expected_vrf_default_routes)
            LOG.info(_LI('Syncing Neutron Router DB - creating routers'))
            self.create_routers(routers)
            LOG.info(_LI('Syncing Neutron Router DB - creating interfaces'))
            self.create_router_interfaces(router_interfaces)
            LOG.info(_LI('Syncing Neutron Router DB finished'))
        except Exception:
            exc_str = traceback.format_exc()
            LOG.error(_LE("Error during synchronize processing %s"), exc_str)

    def get_vrfs(self, server):
        ret = self.driver._run_eos_cmds(['show vrf'], server)
        if len(ret or []) != 1 or 'vrfs' not in ret[0].keys():
            return set()
        eos_vrfs = set(vrf for vrf in ret[0]['vrfs'].keys()
                       if vrf.startswith('__OpenStack__'))
        return eos_vrfs

    def get_svis(self, server):
        ret = self.driver._run_eos_cmds(['show ip interface'], server)
        if len(ret or []) != 1 or 'interfaces' not in ret[0].keys():
            return set()
        eos_svis = set(
            int(vlan.strip('Vlan'))
            for vlan in ret[0]['interfaces'].keys() if 'Vlan' in vlan)
        return eos_svis

    def get_vlans(self, server):
        ret = self.driver._run_eos_cmds(['show vlan'], server)
        if len(ret or []) != 1 or 'vlans' not in ret[0].keys():
            return set()
        eos_vlans = set(int(vlan) for vlan, info in ret[0]['vlans'].items()
                        if not info['dynamic'])
        return eos_vlans

    def _get_vrf_default_route(self, vrf, vrf_routes):
        if (not vrf_routes or
                'routes' not in vrf_routes or
                '0.0.0.0/0' not in vrf_routes['routes'] or
                'vias' not in vrf_routes['routes']['0.0.0.0/0']):
            return set()
        result = set()
        for via in vrf_routes['routes']['0.0.0.0/0']['vias']:
            result.add((vrf, via.get('nexthopAddr')))
        return result

    def get_vrf_default_routes(self, server):
        ret = self.driver._run_eos_cmds(['show ip route vrf all'], server)
        if len(ret or []) != 1 or 'vrfs' not in ret[0].keys():
            return set()

        eos_vrf_gateway = set()
        for vrf, vrf_route in ret[0]['vrfs'].items():
            if '__OpenStack__' in vrf:
                eos_vrf_gateway.update(self._get_vrf_default_route(vrf,
                                                                   vrf_route))
        return eos_vrf_gateway

    def do_cleanup(self, expected_vrfs, expected_vlans,
                   expected_vrf_default_routes):
        for server in self._servers:
            eos_svis = self.get_svis(server)
            eos_vlans = self.get_vlans(server)
            svis_to_delete = (eos_svis - self._protected_vlans -
                              expected_vlans)
            vlans_to_delete = (eos_vlans - self._protected_vlans -
                               expected_vlans)
            delete_cmds = []
            delete_cmds.extend('no interface vlan %s' % svi
                               for svi in svis_to_delete)
            delete_cmds.extend('no vlan %s' % vlan
                               for vlan in vlans_to_delete)
            if self._use_vrf:
                eos_vrfs = self.get_vrfs(server)
                vrfs_to_delete = eos_vrfs - expected_vrfs
                delete_cmds.extend([c.format(vrf)
                                    for c in self.driver.routerDict['delete']
                                    for vrf in vrfs_to_delete])
                if self._vrf_default_route:
                    eos_vrf_gateway = self.get_vrf_default_routes(server)

                    vrf_gw_to_delete = (eos_vrf_gateway -
                                        expected_vrf_default_routes)

                    delete_cmds.extend(
                        'no ip route vrf %s 0.0.0.0/0' % vrf
                        for (vrf, _) in vrf_gw_to_delete)
            if delete_cmds:
                self.driver._run_config_cmds(delete_cmds, server)

    def create_routers(self, routers):
        for r in routers:
            try:
                self.driver.create_router(self, r)
            except Exception:
                LOG.error(_LE("Error Adding router %(router_id)s "
                              "on Arista HW"), {'router_id': r})

    def create_router_interfaces(self, router_interfaces):
        for r in router_interfaces:
            try:
                self.driver.add_router_interface(self, r)
            except Exception:
                LOG.error(_LE("Error Adding interface %(subnet_id)s "
                              "to router %(router_id)s on Arista HW"),
                          {'subnet_id': r['subnet_id'], 'router_id': r['id']})


class AristaL3ServicePlugin(service_base.ServicePluginBase,
                            extraroute_db.ExtraRoute_db_mixin,
                            l3_gwmode_db.L3_NAT_db_mixin,
                            l3_agentschedulers_db.L3AgentSchedulerDbMixin):

    """Implements L3 Router service plugin for Arista hardware.

    Creates routers in Arista hardware, manages them, adds/deletes interfaces
    to the routes.
    """

    supported_extension_aliases = ["router", "ext-gw-mode",
                                   "extraroute"]

    def __init__(self, driver=None):
        super(AristaL3ServicePlugin, self).__init__()
        self.driver = driver or arista_l3_driver.AristaL3Driver()
        self.setup_rpc()
        self.add_worker(AristaL3SyncWorker(self.driver))

    def setup_rpc(self):
        # RPC support
        self.topic = topics.L3PLUGIN
        self.conn = n_rpc.Connection()
        self.agent_notifiers.update(
            {n_const.AGENT_TYPE_L3: l3_rpc_agent_api.L3AgentNotifyAPI()})
        self.endpoints = [l3_rpc.L3RpcCallback()]
        self.conn.create_consumer(self.topic, self.endpoints,
                                  fanout=False)
        self.conn.consume_in_threads()

    def get_plugin_type(self):
        return plugin_constants.L3

    def get_plugin_description(self):
        """Returns string description of the plugin."""
        return ("Arista L3 Router Service Plugin for Arista Hardware "
                "based routing")

    @log_helpers.log_method_call
    def create_router(self, context, router):
        """Create a new router entry in DB, and create it Arista HW."""

        # Add router to the DB
        new_router = super(AristaL3ServicePlugin, self).create_router(
            context,
            router)
        # create router on the Arista Hw
        try:
            self.driver.create_router(context, new_router)
            return new_router
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Error creating router on Arista HW router=%s "),
                          new_router)
                super(AristaL3ServicePlugin, self).delete_router(
                    context,
                    new_router['id']
                )

    @log_helpers.log_method_call
    def update_router(self, context, router_id, router):
        """Update an existing router in DB, and update it in Arista HW."""

        # Read existing router record from DB
        original_router = self.get_router(context, router_id)
        # Update router DB
        new_router = super(AristaL3ServicePlugin, self).update_router(
            context, router_id, router)

        # Modify router on the Arista Hw
        try:
            self.driver.update_router(context, router_id,
                                      original_router, new_router)
            return new_router
        except Exception:
            LOG.error(_LE("Error updating router on Arista HW router=%s "),
                      new_router)

    @log_helpers.log_method_call
    def delete_router(self, context, router_id):
        """Delete an existing router from Arista HW as well as from the DB."""

        router = self.get_router(context, router_id)

        # Delete router on the Arista Hw
        try:
            self.driver.delete_router(context, router_id, router)
        except Exception as e:
            LOG.error(_LE("Error deleting router on Arista HW "
                          "router %(r)s exception=%(e)s"),
                      {'r': router, 'e': e})

        super(AristaL3ServicePlugin, self).delete_router(context, router_id)

    @log_helpers.log_method_call
    def add_router_interface(self, context, router_id, interface_info):
        """Add a subnet of a network to an existing router."""

        new_router = super(AristaL3ServicePlugin, self).add_router_interface(
            context, router_id, interface_info)

        core = directory.get_plugin()

        # Get network info for the subnet that is being added to the router.
        # Check if the interface information is by port-id or subnet-id
        add_by_port, add_by_sub = self._validate_interface_info(interface_info)
        if add_by_sub:
            subnet = core.get_subnet(context, interface_info['subnet_id'])
            # If we add by subnet and we have no port allocated, assigned
            # gateway IP for the interface
            fixed_ip = subnet['gateway_ip']
        elif add_by_port:
            port = core.get_port(context, interface_info['port_id'])
            subnet_id = port['fixed_ips'][0]['subnet_id']
            fixed_ip = port['fixed_ips'][0]['ip_address']
            subnet = core.get_subnet(context, subnet_id)
        network_id = subnet['network_id']

        # To create SVI's in Arista HW, the segmentation Id is required
        # for this network.
        ml2_db = NetworkContext(self, context, {'id': network_id})
        seg_id = ml2_db.network_segments[0]['segmentation_id']

        # Package all the info needed for Hw programming
        router = self.get_router(context, router_id)
        router_info = copy.deepcopy(new_router)
        router_info['seg_id'] = seg_id
        router_info['name'] = router['name']
        router_info['cidr'] = subnet['cidr']
        router_info['gip'] = subnet['gateway_ip']
        router_info['fixed_ip'] = fixed_ip
        router_info['ip_version'] = subnet['ip_version']

        try:
            self.driver.add_router_interface(context, router_info)
            return new_router
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Error Adding subnet %(subnet)s to "
                              "router %(router_id)s on Arista HW"),
                          {'subnet': subnet, 'router_id': router_id})
                super(AristaL3ServicePlugin, self).remove_router_interface(
                    context,
                    router_id,
                    interface_info)

    @log_helpers.log_method_call
    def remove_router_interface(self, context, router_id, interface_info):
        """Remove a subnet of a network from an existing router."""

        router_to_del = (
            super(AristaL3ServicePlugin, self).remove_router_interface(
                context,
                router_id,
                interface_info)
            )

        # Get network information of the subnet that is being removed
        core = directory.get_plugin()
        subnet = core.get_subnet(context, router_to_del['subnet_id'])
        network_id = subnet['network_id']

        # For SVI removal from Arista HW, segmentation ID is needed
        ml2_db = NetworkContext(self, context, {'id': network_id})
        seg_id = ml2_db.network_segments[0]['segmentation_id']

        router = self.get_router(context, router_id)
        router_info = copy.deepcopy(router_to_del)
        router_info['seg_id'] = seg_id
        router_info['name'] = router['name']
        router_info['ip_version'] = subnet['ip_version']
        try:
            self.driver.remove_router_interface(context, router_info)
            return router_to_del
        except Exception as exc:
            LOG.error(_LE("Error removing interface %(interface)s from "
                          "router %(router_id)s on Arista HW"
                          "Exception =(exc)s"),
                      {'interface': interface_info, 'router_id': router_id,
                       'exc': exc})



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eventlet import greenthread
from eventlet import queue
import mock
import time

from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_utils import importutils

from neutron.tests.unit import testlib_api

from networking_arista.common import constants as a_const
from networking_arista.common import exceptions as arista_exc
from networking_arista.ml2 import arista_sync
from networking_arista.ml2.mechanism_arista import MechResource
from networking_arista.tests.unit import utils


class SyncServiceTest(testlib_api.SqlTestCase):
    """Test cases for the sync service."""

    def setUp(self):
        super(SyncServiceTest, self).setUp()
        utils.setup_arista_wrapper_config(cfg)
        plugin_klass = importutils.import_class(
            "neutron.db.db_base_plugin_v2.NeutronDbPluginV2")
        directory.add_plugin(plugin_constants.CORE, plugin_klass())
        utils.setup_scenario()
        self.mech_queue = queue.LightQueue()
        self.sync_service = arista_sync.AristaSyncWorker(self.mech_queue)
        self.sync_service._rpc = utils.MockCvx('region')

    def tearDown(self):
        if self.sync_service._running:
            self.sync_service.stop()
            self.sync_service.wait()
        super(SyncServiceTest, self).tearDown()

    def test_start(self):
        self.sync_service.start()
        self.assertTrue(self.sync_service._running)
        self.assertIsNotNone(self.sync_service._thread)
        self.assertIsNotNone(self.sync_service.sync_order)
        self.assertIsNotNone(self.sync_service.done)

    def test_start_twice(self):
        self.sync_service.start()
        current_thread = self.sync_service._thread
        self.sync_service.start()
        self.assertEqual(self.sync_service._thread, current_thread)

    def test_stop_graceful(self):
        self.test_start()
        running_thread = self.sync_service._thread
        self.sync_service.stop()
        self.sync_service.wait()
        self.assertFalse(self.sync_service._running)
        self.assertTrue(running_thread.dead)
        self.assertIsNone(self.sync_service._thread)

    def test_stop_ungraceful(self):
        self.test_start()
        running_thread = self.sync_service._thread
        self.sync_service.stop(graceful=False)
        self.assertTrue(running_thread.dead)
        self.assertIsNone(self.sync_service._thread)

    def test_reset(self):
        self.test_start()
        old_thread = self.sync_service._thread
        self.sync_service.reset()
        self.assertNotEqual(self.sync_service._thread, old_thread)

    def test_resource_class_full_coverage(self):
        self.sync_service.initialize()
        for i in range(len(self.sync_service.sync_order)):
            self.sync_service.sync_order[i] = mock.MagicMock()
        self.sync_service.synchronize_resources()
        for resource_type in self.sync_service.sync_order:
            resource_type.delete_cvx_resources.assert_called_once()
            resource_type.create_cvx_resources.assert_called_once()

    def test_full_sync_cvx_populated(self):
        self.sync_service.initialize()
        self.sync_service.synchronize_resources()
        for endpoint, data in self.sync_service._rpc.endpoint_data.items():
            self.assertNotEqual(data, {})

    def test_process_mech_update(self):
        self.sync_service.initialize()
        for resource_type in a_const.ALL_RESOURCE_TYPES:
            res_cls = mock.MagicMock()
            with mock.patch.object(self.sync_service,
                                   'get_resource_class') as get:
                get.return_value = res_cls
                res = MechResource('id', resource_type, a_const.CREATE)
                self.sync_service.update_neutron_resource(res)
                get.assert_called_once_with(resource_type)
                res_cls.update_neutron_resource.assert_called_once_with(
                    'id', a_const.CREATE)
                get.reset_mock()
                get.return_value = res_cls
                res = MechResource('id', resource_type, a_const.DELETE)
                self.sync_service.update_neutron_resource(res)
                get.assert_called_once_with(resource_type)
                res_cls.update_neutron_resource.assert_called_once_with(
                    'id', a_const.DELETE)

    def test_force_full_sync(self):
        self.sync_service.initialize()
        for i in range(len(self.sync_service.sync_order)):
            self.sync_service.sync_order[i] = mock.MagicMock()
        self.sync_service.force_full_sync()
        for resource_type in self.sync_service.sync_order:
            resource_type.clear_all_data.assert_called_once()

    def test_sync_timeout(self):
        self.sync_service.initialize()
        with mock.patch.object(
                self.sync_service, 'check_if_out_of_sync') as oos:
            self.sync_service.wait_for_sync_required()
            oos.assert_called_once()

    def test_full_sync_required(self):
        self.sync_service.initialize()
        self.sync_service._cvx_uuid = 'old-id'
        self.sync_service._rpc = mock.MagicMock()
        self.sync_service._rpc.get_cvx_uuid.return_value = 'new-id'
        with mock.patch.object(self.sync_service, 'force_full_sync') as ffs:
            self.assertTrue(self.sync_service.check_if_out_of_sync())
            ffs.assert_called_once()
            self.assertEqual(self.sync_service._synchronizing_uuid, 'new-id')
            self.assertNotEqual(self.sync_service._last_sync_time, 0)

    def test_mech_queue_timeout(self):
        self.sync_service.initialize()
        self.assertFalse(self.sync_service.wait_for_mech_driver_update(1))

    def test_mech_queue_updated(self):
        self.sync_service.initialize()
        resource = MechResource('tid', a_const.TENANT_RESOURCE, a_const.CREATE)
        self.mech_queue.put(resource)
        # Must yield to allow resource to be available on the queue
        greenthread.sleep(0)
        self.assertTrue(self.sync_service.wait_for_mech_driver_update(1))
        self.assertEqual(self.sync_service._resources_to_update, [resource])

    def test_sync_start_fail(self):
        self.sync_service.initialize()
        self.sync_service._rpc = mock.MagicMock()
        self.sync_service._rpc.sync_start.return_value = False
        self.assertEqual(self.sync_service._last_sync_time, 0)
        for i in range(len(self.sync_service.sync_order)):
            self.sync_service.sync_order[i] = mock.MagicMock()
        self.sync_service.synchronize_resources()
        for resource_type in self.sync_service.sync_order:
            resource_type.delete_cvx_resources.assert_not_called()
            resource_type.create_cvx_resources.assert_not_called()
        self.sync_service._rpc.sync_end.assert_not_called()

    def test_sync_lock_release_on_failure(self):
        self.sync_service.initialize()
        self.sync_service._rpc = mock.MagicMock()
        self.sync_service._rpc.get_cvx_uuid.return_value = 'cvx-id'
        self.sync_service.update_neutron_resource = mock.Mock(
            side_effect=arista_exc.AristaRpcError(msg='fail'))
        resource = MechResource('tid', a_const.TENANT_RESOURCE, a_const.CREATE)
        self.sync_service.start()
        self.sync_service._last_sync_time = time.time()
        self.sync_service._cvx_uuid = 'cvx-id'
        self.sync_service._synchronizing_uuid = 'cvx-id'
        self.mech_queue.put(resource)
        greenthread.sleep(0)
        self.sync_service.stop()
        self.sync_service.wait()
        self.sync_service._rpc.sync_end.assert_called_once()

    def test_failure_after_switchover(self):
        """Test behaviour of sync fail after CVX switchover

        On sync failure, we should retain old sync UUID and we should not set
        sync complete on CVX to avoid publishing config based on incomplete
        state.
        """
        self.sync_service.initialize()
        self.sync_service._rpc = mock.MagicMock()
        self.sync_service._rpc.get_cvx_uuid.return_value = 'new-id'
        self.sync_service.start()
        self.sync_service._cvx_uuid = 'old-id'
        self.sync_service.tenants.delete_cvx_resources = mock.Mock(
            side_effect=arista_exc.AristaRpcError(msg='fail'))
        greenthread.sleep(0)
        self.sync_service.stop()
        self.sync_service.wait()
        self.sync_service._rpc.sync_start.assert_called()
        self.sync_service._rpc.sync_end.assert_not_called()
        self.assertEqual(self.sync_service._cvx_uuid, 'old-id')
        self.assertIsNone(self.sync_service._synchronizing_uuid, 'new-id')



# Copyright 2017 Arista Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from alembic import op

"""Drop AristaProvisionedVms

Revision ID: 941bad5630c1
Revises: 39c2eeb67116
Create Date: 2017-08-31 17:44:21.334780

"""

# revision identifiers, used by Alembic.
revision = '941bad5630c1'
down_revision = '39c2eeb67116'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table('arista_provisioned_vms')



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from oslo_config import cfg
from oslo_log import log as logging
from sqlalchemy import and_, or_
from sqlalchemy import func
from sqlalchemy.orm import immediateload, Query, aliased

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const
from neutron_lib.db import api as db
from neutron_lib.services.trunk import constants as t_const

from neutron.db.models import l3 as l3_models
from neutron.db.models import l3ha as l3ha_models
from neutron.db.models.plugins.ml2 import vlanallocation
from neutron.db.models import securitygroup as sg_models
from neutron.db.models import segment as segment_models
from neutron.db import models_v2
from neutron.plugins.ml2 import models as ml2_models
from neutron.services.trunk import models as trunk_models

from networking_arista.common import utils

LOG = logging.getLogger(__name__)


def join_if_necessary(query, *args, **kwargs):
    table = args[0]
    if table in [t.entity for t in query._join_entities]:
        return query
    elif table in query._primary_entity.entities:
        return query
    return query.join(*args, **kwargs)


def outerjoin_if_necessary(query, *args, **kwargs):
    table = args[0]
    if table in [t.entity for t in query._join_entities]:
        return query
    elif table in query._primary_entity.entities:
        return query
    return query.outerjoin(*args, **kwargs)


def filter_network_type(query):
    """Filter unsupported segment types"""
    segment_model = segment_models.NetworkSegment
    query = (query
             .filter(
                 segment_model.network_type.in_(
                     utils.SUPPORTED_NETWORK_TYPES)))
    return query


def filter_unbound_ports(query):
    """Filter ports not bound to a host or network"""
    # hack for pep8 E711: comparison to None should be
    # 'if cond is not None'
    none = None
    port_model = models_v2.Port
    binding_level_model = ml2_models.PortBindingLevel
    query = (query
             .join_if_necessary(port_model)
             .join_if_necessary(binding_level_model)
             .filter(
                 binding_level_model.host != '',
                 port_model.device_id != none,
                 port_model.network_id != none))
    return query


def filter_by_device_owner(query, device_owners=None):
    """Filter ports by device_owner

    Either filter using specified device_owner or using the list of all
    device_owners supported and unsupported by the arista ML2 plugin
    """
    port_model = models_v2.Port
    if not device_owners:
        device_owners = utils.SUPPORTED_DEVICE_OWNERS
    supported_device_owner_filter = [
        port_model.device_owner.ilike('%s%%' % owner)
        for owner in device_owners]
    unsupported_device_owner_filter = [
        port_model.device_owner.notilike('%s%%' % owner)
        for owner in utils.UNSUPPORTED_DEVICE_OWNERS]
    query = (query
             .filter(
                 and_(*unsupported_device_owner_filter),
                 or_(*supported_device_owner_filter)))
    return query


def filter_by_device_id(query):
    """Filter ports attached to devices we don't care about

    Currently used to filter DHCP_RESERVED ports
    """
    port_model = models_v2.Port
    unsupported_device_id_filter = [
        port_model.device_id.notilike('%s%%' % id)
        for id in utils.UNSUPPORTED_DEVICE_IDS]
    query = (query
             .filter(and_(*unsupported_device_id_filter)))
    return query


def filter_by_vnic_type(query, vnic_type):
    """Filter ports by vnic_type (currently only used for baremetals)"""
    port_model = models_v2.Port
    binding_model = ml2_models.PortBinding
    dst_binding_model = ml2_models.DistributedPortBinding
    query = (query
             .outerjoin_if_necessary(
                 binding_model,
                 port_model.id == binding_model.port_id)
             .outerjoin_if_necessary(
                 dst_binding_model,
                 port_model.id == dst_binding_model.port_id)
             .filter(
                 (binding_model.vnic_type == vnic_type) |
                 (dst_binding_model.vnic_type == vnic_type)))
    return query


def filter_unmanaged_physnets(query):
    """Filter ports managed by other ML2 plugins """
    config = cfg.CONF.ml2_arista
    managed_physnets = config['managed_physnets']

    # Filter out ports bound to segments on physnets that we're not
    # managing
    segment_model = segment_models.NetworkSegment
    if managed_physnets:
        query = (query
                 .join_if_necessary(segment_model)
                 .filter(segment_model.physical_network.in_(
                     managed_physnets)))
    return query


def filter_inactive_ports(query):
    """Filter ports that aren't in active status """
    port_model = models_v2.Port
    query = (query
             .filter(port_model.status.in_([n_const.PORT_STATUS_ACTIVE,
                                            n_const.PORT_STATUS_BUILD])))
    return query


def filter_unnecessary_ports(query, device_owners=None, vnic_type=None,
                             active=True):
    """Filter out all ports are not needed on CVX """
    query = (query
             .filter_unbound_ports()
             .filter_by_device_owner(device_owners)
             .filter_by_device_id()
             .filter_unmanaged_physnets())
    if active:
        query = query.filter_inactive_ports()
    if vnic_type:
        query = query.filter_by_vnic_type(vnic_type)
    return query


Query.join_if_necessary = join_if_necessary
Query.outerjoin_if_necessary = outerjoin_if_necessary
Query.filter_network_type = filter_network_type
Query.filter_unbound_ports = filter_unbound_ports
Query.filter_by_device_owner = filter_by_device_owner
Query.filter_by_device_id = filter_by_device_id
Query.filter_by_vnic_type = filter_by_vnic_type
Query.filter_unmanaged_physnets = filter_unmanaged_physnets
Query.filter_inactive_ports = filter_inactive_ports
Query.filter_unnecessary_ports = filter_unnecessary_ports


def get_tenants(tenant_id=None):
    """Returns list of all project/tenant ids that may be relevant on CVX"""
    session = db.get_reader_session()
    project_ids = set()
    with session.begin():
        for m in [models_v2.Network, models_v2.Port, l3_models.Router]:
            q = session.query(m.project_id).filter(m.project_id != '')
            if tenant_id is not None:
                q = q.filter(m.project_id == tenant_id)
            project_ids.update(pid[0] for pid in q.distinct())
    return [{'project_id': project_id} for project_id in project_ids]


def get_networks(network_id=None):
    """Returns list of all networks that may be relevant on CVX"""
    session = db.get_reader_session()
    with session.begin():
        model = models_v2.Network
        l3ha_network = l3ha_models.L3HARouterNetwork
        networks = (session.query(model, l3ha_network)
                           .outerjoin(l3ha_network,
                                      l3ha_network.network_id == model.id))
        if network_id is not None:
            networks = networks.filter(model.id == network_id)
    return networks.all()


def get_segments(segment_id=None):
    """Returns list of all network segments that may be relevant on CVX"""
    session = db.get_reader_session()
    with session.begin():
        model = segment_models.NetworkSegment
        segments = session.query(model).filter_network_type()
        if segment_id is not None:
            segments = segments.filter(model.id == segment_id)
    return segments.all()


def get_instances(device_owners=None, vnic_type=None, instance_id=None):
    """Returns filtered list of all instances in the neutron db"""
    session = db.get_reader_session()
    with session.begin():
        port_model = models_v2.Port
        binding_model = ml2_models.PortBinding
        router_model = l3_models.Router
        instances = (session
                     .query(port_model,
                            binding_model,
                            router_model)
                     .outerjoin(
                         binding_model,
                         port_model.id == binding_model.port_id)
                     .outerjoin(
                         router_model,
                         port_model.device_id == router_model.id)
                     .distinct(port_model.device_id)
                     .group_by(port_model.device_id)
                     .filter_unnecessary_ports(device_owners, vnic_type))
        if instance_id is not None:
            instances = instances.filter(port_model.device_id == instance_id)
    return instances.all()


def get_dhcp_instances(instance_id=None):
    """Returns filtered list of DHCP instances that may be relevant on CVX"""
    return get_instances(device_owners=[n_const.DEVICE_OWNER_DHCP],
                         instance_id=instance_id)


def get_router_instances(instance_id=None):
    """Returns filtered list of routers that may be relevant on CVX"""
    return get_instances(device_owners=[n_const.DEVICE_OWNER_DVR_INTERFACE,
                                        n_const.DEVICE_OWNER_ROUTER_HA_INTF,
                                        n_const.DEVICE_OWNER_ROUTER_INTF,
                                        n_const.DEVICE_OWNER_ROUTER_GW],
                         instance_id=instance_id)


def get_vm_instances(instance_id=None):
    """Returns filtered list of vms that may be relevant on CVX"""
    return get_instances(device_owners=[n_const.DEVICE_OWNER_COMPUTE_PREFIX],
                         vnic_type=portbindings.VNIC_NORMAL,
                         instance_id=instance_id)


def get_baremetal_instances(instance_id=None):
    """Returns filtered list of baremetals that may be relevant on CVX"""
    return get_instances(vnic_type=portbindings.VNIC_BAREMETAL,
                         instance_id=instance_id)


def get_ports(device_owners=None, vnic_type=None, port_id=None, active=True):
    """Returns list of all ports in neutron the db"""
    session = db.get_reader_session()
    with session.begin():
        port_model = models_v2.Port
        router_model = l3_models.Router
        ports = (session
                 .query(port_model, router_model)
                 .outerjoin(router_model,
                            router_model.id == port_model.device_id)
                 .filter_unnecessary_ports(device_owners, vnic_type, active))
        if port_id is not None:
            ports = ports.filter(port_model.id == port_id)
    return ports.all()


def get_dhcp_ports(port_id=None):
    """Returns filtered list of DHCP instances that may be relevant on CVX"""
    return get_ports(device_owners=[n_const.DEVICE_OWNER_DHCP],
                     port_id=port_id)


def get_router_ports(port_id=None):
    """Returns filtered list of routers that may be relevant on CVX"""
    return get_ports(device_owners=[n_const.DEVICE_OWNER_DVR_INTERFACE,
                                    n_const.DEVICE_OWNER_ROUTER_HA_INTF,
                                    n_const.DEVICE_OWNER_ROUTER_INTF,
                                    n_const.DEVICE_OWNER_ROUTER_GW],
                     port_id=port_id)


def get_vm_ports(port_id=None):
    """Returns filtered list of vms that may be relevant on CVX"""
    return get_ports(device_owners=[n_const.DEVICE_OWNER_COMPUTE_PREFIX,
                                    t_const.TRUNK_SUBPORT_OWNER],
                     vnic_type=portbindings.VNIC_NORMAL, port_id=port_id)


def get_baremetal_ports(port_id=None):
    """Returns filtered list of baremetals that may be relevant on CVX"""
    return get_ports(vnic_type=portbindings.VNIC_BAREMETAL, port_id=port_id)


def get_port_bindings(binding_key=None):
    """Returns filtered list of port bindings that may be relevant on CVX

    This query is a little complex as we need all binding levels for any
    binding that has a single managed physnet, but we need to filter bindings
    that have no managed physnets. In order to achieve this, we join to the
    binding_level_model once to filter bindings with no managed levels,
    then a second time to get all levels for the remaining bindings.

    The loop at the end is a convenience to associate levels with bindings
    as a list. This would ideally be done through the use of an orm.relation,
    but due to some sqlalchemy limitations imposed to make OVO work, we can't
    add relations to existing models.
    """
    session = db.get_reader_session()
    with session.begin():
        binding_level_model = ml2_models.PortBindingLevel
        aliased_blm = aliased(ml2_models.PortBindingLevel)
        port_binding_model = ml2_models.PortBinding
        dist_binding_model = ml2_models.DistributedPortBinding
        bindings = (session.query(port_binding_model, aliased_blm)
                    .join(binding_level_model,
                          and_(
                              port_binding_model.port_id ==
                              binding_level_model.port_id,
                              port_binding_model.host ==
                              binding_level_model.host))
                    .filter_unnecessary_ports()
                    .join(aliased_blm,
                          and_(port_binding_model.port_id ==
                               aliased_blm.port_id,
                               port_binding_model.host ==
                               aliased_blm.host)))
        dist_bindings = (session.query(dist_binding_model, aliased_blm)
                         .join(
                             binding_level_model,
                             and_(dist_binding_model.port_id ==
                                  binding_level_model.port_id,
                                  dist_binding_model.host ==
                                  binding_level_model.host))
                         .filter_unnecessary_ports()
                         .filter(dist_binding_model.status.in_([
                                 n_const.PORT_STATUS_BUILD,
                                 n_const.PORT_STATUS_ACTIVE]))
                         .join(aliased_blm,
                               and_(dist_binding_model.port_id ==
                                    aliased_blm.port_id,
                                    dist_binding_model.host ==
                                    aliased_blm.host)))
        if binding_key:
            port_id = binding_key[0]
            if type(binding_key[1]) == tuple:
                switch_id = binding_key[1][0]
                switch_port = binding_key[1][1]
                bindings = bindings.filter(and_(
                    port_binding_model.port_id == port_id,
                    port_binding_model.profile.ilike('%%%s%%' % switch_id),
                    port_binding_model.profile.ilike('%%%s%%' % switch_port)))
                dist_bindings = dist_bindings.filter(and_(
                    dist_binding_model.port_id == port_id,
                    dist_binding_model.profile.ilike('%%%s%%' % switch_id),
                    dist_binding_model.profile.ilike('%%%s%%' % switch_port)))
            else:
                host_id = binding_key[1]
                bindings = bindings.filter(and_(
                    port_binding_model.port_id == port_id,
                    port_binding_model.host == host_id))
                dist_bindings = dist_bindings.filter(and_(
                    dist_binding_model.port_id == port_id,
                    dist_binding_model.host == host_id))
    binding_levels = collections.defaultdict(list)
    for binding, level in bindings.all() + dist_bindings.all():
        binding_levels[binding].append(level)
    bindings_with_levels = list()
    for binding, levels in binding_levels.items():
        binding.levels = levels
        bindings_with_levels.append(binding)
    return bindings_with_levels


def get_mlag_physnets():
    mlag_pairs = dict()
    session = db.get_reader_session()
    with session.begin():
        physnets = session.query(
            vlanallocation.VlanAllocation.physical_network
        ).distinct().all()
        for (physnet,) in physnets:
            if '_' in physnet:
                peers = physnet.split('_')
                mlag_pairs[peers[0]] = physnet
                mlag_pairs[peers[1]] = physnet
    return mlag_pairs


def segment_is_dynamic(segment_id):
    session = db.get_reader_session()
    with session.begin():
        segment_model = segment_models.NetworkSegment
        res = bool(session
                   .query(segment_model)
                   .filter_by(id=segment_id)
                   .filter_by(is_dynamic=True).count())
    return res


def segment_bound(segment_id):
    session = db.get_reader_session()
    with session.begin():
        binding_level_model = ml2_models.PortBindingLevel
        res = bool(session
                   .query(binding_level_model)
                   .filter_by(segment_id=segment_id).count())
    return res


def tenant_provisioned(tenant_id):
    """Returns true if any networks or ports exist for a tenant."""
    session = db.get_reader_session()
    with session.begin():
        res = any(
            session.query(m).filter(m.tenant_id == tenant_id).count()
            for m in [models_v2.Network, models_v2.Port]
        )
    return res


def instance_provisioned(device_id):
    """Returns true if any ports exist for an instance."""
    session = db.get_reader_session()
    with session.begin():
        port_model = models_v2.Port
        res = bool(session.query(port_model)
                   .filter(port_model.device_id == device_id)
                   .filter_inactive_ports().count())
    return res


def port_provisioned(port_id):
    """Returns true if port still exists."""
    session = db.get_reader_session()
    with session.begin():
        port_model = models_v2.Port
        res = bool(session.query(port_model)
                   .filter(port_model.id == port_id)
                   .filter_inactive_ports().count())
    return res


def get_parent(port_id):
    """Get trunk subport's parent port"""
    session = db.get_reader_session()
    res = dict()
    with session.begin():
        subport_model = trunk_models.SubPort
        trunk_model = trunk_models.Trunk
        subport = (session.query(subport_model).
                   filter(subport_model.port_id == port_id).first())
        if subport:
            trunk = (session.query(trunk_model).
                     filter(trunk_model.id == subport.trunk_id).first())
            if trunk:
                trunk_port_id = trunk.port.id
                res = getattr(get_ports(port_id=trunk_port_id,
                                        active=False)[0], 'Port')
    return res


def get_port_binding_level(filters):
    """Returns entries from PortBindingLevel based on the specified filters."""
    session = db.get_reader_session()
    with session.begin():
        return (session.query(ml2_models.PortBindingLevel).
                filter_by(**filters).
                order_by(ml2_models.PortBindingLevel.level).
                all())


def get_security_groups():
    session = db.get_reader_session()
    with session.begin():
        sg_model = sg_models.SecurityGroup
        # We do an immediate load to prevent the need for the sync worker
        # to issue subqueries
        security_groups = (session.query(sg_model)
                           .options(immediateload(sg_model.rules)))
    return security_groups


def get_baremetal_sg_bindings():
    session = db.get_reader_session()
    with session.begin():
        sg_binding_model = sg_models.SecurityGroupPortBinding
        binding_model = ml2_models.PortBinding
        sg_bindings = (session
                       .query(sg_binding_model,
                              binding_model)
                       .outerjoin(
                           binding_model,
                           sg_binding_model.port_id == binding_model.port_id)
                       .filter_unnecessary_ports(
                           vnic_type=portbindings.VNIC_BAREMETAL)
                       .group_by(sg_binding_model.port_id)
                       .having(func.count(sg_binding_model.port_id) == 1))
    return sg_bindings


def get_subnet_gateway_ipv4(subnet_id):
    """Returns the IPv4 gateway of the router if exists"""
    session = db.get_reader_session()
    with session.begin():
        subnet_model = models_v2.Subnet
        router_model = l3_models.Router
        port_model = models_v2.Port
        ip_allocation_model = models_v2.IPAllocation
        result = (session
                  .query(router_model.name,
                         router_model.id,
                         subnet_model.gateway_ip.label('gip'),
                         subnet_model.cidr,
                         subnet_model.ip_version,
                         port_model.network_id,
                         ip_allocation_model.ip_address)
                  .filter(and_(router_model.gw_port_id == port_model.id,
                               port_model.network_id ==
                               subnet_model.network_id,
                               ip_allocation_model.network_id ==
                               subnet_model.network_id,
                               subnet_model.ip_version == 4,
                               subnet_model.id == subnet_id)).first())
    return _format_gateway_result(result)


def get_network_gateway_ipv4(network_id):
    """Returns all the routers and IPv4 gateway that have network as gateway"""
    session = db.get_reader_session()
    with session.begin():
        subnet_model = models_v2.Subnet
        router_model = l3_models.Router
        port_model = models_v2.Port
        ip_allocation_model = models_v2.IPAllocation
        result = (session
                  .query(router_model.name,
                         router_model.id,
                         subnet_model.gateway_ip.label('gip'),
                         subnet_model.cidr,
                         subnet_model.ip_version,
                         port_model.network_id,
                         ip_allocation_model.ip_address)
                  .filter(and_(router_model.gw_port_id == port_model.id,
                               port_model.network_id ==
                               subnet_model.network_id,
                               ip_allocation_model.network_id ==
                               subnet_model.network_id,
                               port_model.network_id == network_id,
                               subnet_model.ip_version == 4)).first())
    return _format_gateway_result(result)


def _format_gateway_result(db_result):
    """This function formats result as needed by add_router_interface"""
    if not db_result:
        return None
    result = {
        k: db_result[i]
        for i, k in enumerate(
            ('name', 'id', 'gip', 'cidr', 'ip_version', 'network_id',
             'fixed_ip'))}
    return result
